from __future__ import annotations

import logging
import pathlib
import re

import pytest

from great_expectations.experimental.datasources.data_asset.data_connector.regex_parser import (
    RegExParser,
)

logger = logging.getLogger(__file__)


@pytest.fixture
def csv_path() -> pathlib.Path:
    relative_path = pathlib.Path(
        "..",
        "..",
        "..",
        "..",
        "test_sets",
        "taxi_yellow_tripdata_samples",
        "yellow_tripdata_sample_2020-03.csv",
    )
    abs_csv_path = (
        pathlib.Path(__file__).parent.joinpath(relative_path).resolve(strict=True)
    )
    return abs_csv_path


@pytest.fixture
def regex_pattern_two_named_groups() -> re.Pattern:
    return re.compile(r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv")


@pytest.fixture
def regex_pattern_first_named_group_second_common_group() -> re.Pattern:
    return re.compile(r"yellow_tripdata_sample_(?P<year>\d{4})-(\d{2})\.csv")


@pytest.fixture
def regex_pattern_first_common_group_second_named_group() -> re.Pattern:
    return re.compile(r"yellow_tripdata_sample_(\d{4})-(?P<month>\d{2})\.csv")


@pytest.fixture
def regex_pattern_two_common_groups() -> re.Pattern:
    return re.compile(r"yellow_tripdata_sample_(\d{4})-(\d{2})\.csv")


@pytest.fixture
def regex_pattern_no_groups() -> re.Pattern:
    return re.compile(r"yellow_tripdata_sample_\d{4}-\d{2}\.csv")


@pytest.mark.unit
def test_regex_pattern_two_named_groups(
    regex_pattern_two_named_groups: re.Pattern, csv_path: pathlib.Path
):
    regex_pattern: re.Pattern = regex_pattern_two_named_groups
    target: str = csv_path.name
    regex_parser = RegExParser(
        regex_pattern=regex_pattern, unnamed_regex_group_prefix="batch_request_param_"
    )

    assert regex_parser.get_num_all_matched_group_values() == 2
    assert regex_parser.get_num_named_matched_group_values() == 2
    assert regex_parser.get_named_group_name_to_group_index_mapping() == {
        "year": 1,
        "month": 2,
    }
    assert regex_parser.get_named_group_name_to_group_value_mapping(target=target) == {
        "year": "2020",
        "month": "03",
    }
    assert regex_parser.get_all_matched_group_values(target=target) == ["2020", "03"]
    assert (
        regex_parser.get_all_group_names_to_group_indexes_bidirectional_mappings()
        == ({"year": 1, "month": 2}, {1: "year", 2: "month"})
    )
    assert regex_parser.get_all_group_name_to_group_index_mapping() == {
        "year": 1,
        "month": 2,
    }
    assert regex_parser.get_all_group_index_to_group_name_mapping() == {
        1: "year",
        2: "month",
    }
    assert regex_parser.get_all_group_names() == ["year", "month"]
    assert regex_parser.get_all_group_indexes() == [1, 2]
    assert regex_parser.get_group_name_to_group_value_mapping(target=target) == {
        "year": "2020",
        "month": "03",
    }
    assert regex_parser.get_group_index_to_group_value_mapping(target=target) == {
        1: "2020",
        2: "03",
    }


@pytest.mark.unit
def test_regex_pattern_first_named_group_second_common_group(
    regex_pattern_first_named_group_second_common_group: re.Pattern,
    csv_path: pathlib.Path,
):
    regex_pattern: re.Pattern = regex_pattern_first_named_group_second_common_group
    target: str = csv_path.name
    regex_parser = RegExParser(
        regex_pattern=regex_pattern, unnamed_regex_group_prefix="batch_request_param_"
    )

    assert regex_parser.get_num_all_matched_group_values() == 2
    assert regex_parser.get_num_named_matched_group_values() == 1
    assert regex_parser.get_named_group_name_to_group_index_mapping() == {"year": 1}
    assert regex_parser.get_named_group_name_to_group_value_mapping(target=target) == {
        "year": "2020"
    }
    assert regex_parser.get_all_matched_group_values(target=target) == ["2020", "03"]
    assert (
        regex_parser.get_all_group_names_to_group_indexes_bidirectional_mappings()
        == (
            {"year": 1, "batch_request_param_2": 2},
            {1: "year", 2: "batch_request_param_2"},
        )
    )
    assert regex_parser.get_all_group_name_to_group_index_mapping() == {
        "year": 1,
        "batch_request_param_2": 2,
    }
    assert regex_parser.get_all_group_index_to_group_name_mapping() == {
        1: "year",
        2: "batch_request_param_2",
    }
    assert regex_parser.get_all_group_names() == ["year", "batch_request_param_2"]
    assert regex_parser.get_all_group_indexes() == [1, 2]
    assert regex_parser.get_group_name_to_group_value_mapping(target=target) == {
        "year": "2020",
        "batch_request_param_2": "03",
    }
    assert regex_parser.get_group_index_to_group_value_mapping(target=target) == {
        1: "2020",
        2: "03",
    }


@pytest.mark.unit
def test_regex_pattern_first_common_group_second_named_group(
    regex_pattern_first_common_group_second_named_group: re.Pattern,
    csv_path: pathlib.Path,
):
    regex_pattern: re.Pattern = regex_pattern_first_common_group_second_named_group
    target: str = csv_path.name
    regex_parser = RegExParser(
        regex_pattern=regex_pattern, unnamed_regex_group_prefix="batch_request_param_"
    )

    assert regex_parser.get_num_all_matched_group_values() == 2
    assert regex_parser.get_num_named_matched_group_values() == 1
    assert regex_parser.get_named_group_name_to_group_index_mapping() == {"month": 2}
    assert regex_parser.get_named_group_name_to_group_value_mapping(target=target) == {
        "month": "03"
    }
    assert regex_parser.get_all_matched_group_values(target=target) == ["2020", "03"]
    assert (
        regex_parser.get_all_group_names_to_group_indexes_bidirectional_mappings()
        == (
            {"batch_request_param_1": 1, "month": 2},
            {1: "batch_request_param_1", 2: "month"},
        )
    )
    assert regex_parser.get_all_group_name_to_group_index_mapping() == {
        "batch_request_param_1": 1,
        "month": 2,
    }
    assert regex_parser.get_all_group_index_to_group_name_mapping() == {
        1: "batch_request_param_1",
        2: "month",
    }
    assert regex_parser.get_all_group_names() == ["batch_request_param_1", "month"]
    assert regex_parser.get_all_group_indexes() == [1, 2]
    assert regex_parser.get_group_name_to_group_value_mapping(target=target) == {
        "batch_request_param_1": "2020",
        "month": "03",
    }
    assert regex_parser.get_group_index_to_group_value_mapping(target=target) == {
        1: "2020",
        2: "03",
    }


@pytest.mark.unit
def test_regex_pattern_two_common_groups(
    regex_pattern_two_common_groups: re.Pattern, csv_path: pathlib.Path
):
    regex_pattern: re.Pattern = regex_pattern_two_common_groups
    target: str = csv_path.name
    regex_parser = RegExParser(
        regex_pattern=regex_pattern, unnamed_regex_group_prefix="batch_request_param_"
    )

    assert regex_parser.get_num_all_matched_group_values() == 2
    assert regex_parser.get_num_named_matched_group_values() == 0
    assert regex_parser.get_named_group_name_to_group_index_mapping() == {}
    assert regex_parser.get_named_group_name_to_group_value_mapping(target=target) == {}
    assert regex_parser.get_all_matched_group_values(target=target) == ["2020", "03"]
    assert (
        regex_parser.get_all_group_names_to_group_indexes_bidirectional_mappings()
        == (
            {"batch_request_param_1": 1, "batch_request_param_2": 2},
            {1: "batch_request_param_1", 2: "batch_request_param_2"},
        )
    )
    assert regex_parser.get_all_group_name_to_group_index_mapping() == {
        "batch_request_param_1": 1,
        "batch_request_param_2": 2,
    }
    assert regex_parser.get_all_group_index_to_group_name_mapping() == {
        1: "batch_request_param_1",
        2: "batch_request_param_2",
    }
    assert regex_parser.get_all_group_names() == [
        "batch_request_param_1",
        "batch_request_param_2",
    ]
    assert regex_parser.get_all_group_indexes() == [1, 2]
    assert regex_parser.get_group_name_to_group_value_mapping(target=target) == {
        "batch_request_param_1": "2020",
        "batch_request_param_2": "03",
    }
    assert regex_parser.get_group_index_to_group_value_mapping(target=target) == {
        1: "2020",
        2: "03",
    }


@pytest.mark.unit
def test_regex_pattern_no_groups(
    regex_pattern_no_groups: re.Pattern, csv_path: pathlib.Path
):
    regex_pattern: re.Pattern = regex_pattern_no_groups
    target: str = csv_path.name
    regex_parser = RegExParser(
        regex_pattern=regex_pattern, unnamed_regex_group_prefix="batch_request_param_"
    )

    assert regex_parser.get_num_all_matched_group_values() == 0
    assert regex_parser.get_num_named_matched_group_values() == 0
    assert regex_parser.get_named_group_name_to_group_index_mapping() == {}
    assert regex_parser.get_named_group_name_to_group_value_mapping(target=target) == {}
    assert regex_parser.get_all_matched_group_values(target=target) == []
    assert (
        regex_parser.get_all_group_names_to_group_indexes_bidirectional_mappings()
        == ({}, {})
    )
    assert regex_parser.get_all_group_name_to_group_index_mapping() == {}
    assert regex_parser.get_all_group_index_to_group_name_mapping() == {}
    assert regex_parser.get_all_group_names() == []
    assert regex_parser.get_all_group_indexes() == []
    assert regex_parser.get_group_name_to_group_value_mapping(target=target) == {}
    assert regex_parser.get_group_index_to_group_value_mapping(target=target) == {}



from __future__ import annotations

from typing import TYPE_CHECKING

import nbformat

from great_expectations.core.expectation_suite import ExpectationSuite  # noqa: TCH001
from great_expectations.dataset import Dataset
from great_expectations.render.renderer.suite_edit_notebook_renderer import (
    SuiteEditNotebookRenderer,
)

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )


class SuiteScaffoldNotebookRenderer(SuiteEditNotebookRenderer):
    def __init__(
        self, context: AbstractDataContext, suite: ExpectationSuite, batch_kwargs
    ) -> None:
        super().__init__(context=context)
        self.suite = suite
        self.suite_name = suite.expectation_suite_name
        self.batch_kwargs = self.get_batch_kwargs(self.suite, batch_kwargs)
        self.batch = self.load_batch()

    def add_header(self) -> None:
        self.add_markdown_cell(
            f"""# Scaffold a new Expectation Suite (Experimental)
This process helps you avoid writing lots of boilerplate when authoring suites by allowing you to select columns and other factors that you care about and letting a profiler write some candidate expectations for you to adjust.

**Expectation Suite Name**: `{self.suite_name}`

We'd love it if you'd **reach out to us on** the [**Great Expectations Slack Channel**](https://greatexpectations.io/slack)!"""
        )

        if not self.batch_kwargs:
            self.batch_kwargs = {}
        self.add_code_cell(
            f"""\
import great_expectations as gx
from great_expectations.checkpoint import LegacyCheckpoint
from great_expectations.profile.user_configurable_profiler import UserConfigurableProfiler
from great_expectations.data_context.types.resource_identifiers import ValidationResultIdentifier

context = gx.get_context()

expectation_suite_name = "{self.suite_name}"

# Wipe the suite clean to prevent unwanted expectations in the batch
suite = context.add_or_update_expectation_suite(expectation_suite_name)

batch_kwargs = {self.batch_kwargs}
batch = context.get_batch(batch_kwargs, suite)
batch.head()""",
            lint=True,
        )

    def _add_scaffold_column_list(self) -> None:
        columns = [f"    '{col}'" for col in self.batch.get_table_columns()]
        columns = ",\n".join(columns)
        code = f"""\
ignored_columns = [
{columns}
]"""
        self.add_code_cell(code, lint=True)

    def add_footer(self) -> None:
        self.add_markdown_cell(
            """## Save & review the scaffolded Expectation Suite

Let's save the scaffolded expectation suite as a JSON file in the
`great_expectations/expectations` directory of your project and rebuild the Data
 Docs site to make it easy to review the scaffolded suite."""
        )
        if self.context and self.context.validation_operators.get(
            "action_list_operator"
        ):
            code_cell = """\
context.update_expectation_suite(expectation_suite=suite)

results = LegacyCheckpoint(
    name="_temp_checkpoint",
    data_context=context,
    batches=[
        {
          "batch_kwargs": batch_kwargs,
          "expectation_suite_names": [expectation_suite_name]
        }
    ],
    validation_operator_name="action_list_operator"
).run()
validation_result_identifier = results.list_validation_result_identifiers()[0]
context.build_data_docs()
context.open_data_docs(validation_result_identifier)"""
        else:
            code_cell = """\
context.update_expectation_suite(expectation_suite=suite)

results = LegacyCheckpoint(
    name="_temp_checkpoint",
    data_context=context,
    batches=[
        {
          "batch_kwargs": batch_kwargs,
          "expectation_suite_names": [expectation_suite_name]
        }
    ]
).run()
validation_result_identifier = results.list_validation_result_identifiers()[0]
context.build_data_docs()
context.open_data_docs(validation_result_identifier)"""
        self.add_code_cell(code_cell)
        self.add_markdown_cell(
            f"""## Next steps
After you review this scaffolded Expectation Suite in Data Docs you
should edit this suite to make finer grained adjustments to the expectations.
This can be done by running `great_expectations suite edit {self.suite_name}`."""
        )

    def load_batch(self):
        batch = self.context.get_batch(self.batch_kwargs, self.suite)
        assert isinstance(
            batch, Dataset
        ), "Batch failed to load. Please check your batch_kwargs"
        return batch

    def render(self, batch_kwargs=None, **kwargs) -> nbformat.NotebookNode:
        self._notebook = nbformat.v4.new_notebook()
        self.add_header()
        self.add_markdown_cell(
            """## Select the columns on which you would like to scaffold expectations and those which you would like to ignore.

Great Expectations will choose which expectations might make sense for a column based on the **data type** and **cardinality** of the data in each selected column.

Simply comment out columns that are important and should be included. You can select multiple lines and
use a jupyter keyboard shortcut to toggle each line: **Linux/Windows**:
`Ctrl-/`, **macOS**: `Cmd-/`"""
        )
        self._add_scaffold_column_list()
        # TODO probably more explanation here about the workflow
        self.add_markdown_cell(
            """## Run the scaffolder

The suites generated here are **not meant to be production suites** - they are **scaffolds to build upon**.

**To get to a production grade suite, you will definitely want to [edit this
suite](https://docs.greatexpectations.io/en/latest/guides/how_to_guides/creating_and_editing_expectations/how_to_edit_an_expectation_suite_using_a_disposable_notebook.html?utm_source=notebook&utm_medium=scaffold_expectations)
after scaffolding gets you close to what you want.**

This is highly configurable depending on your goals.
You can ignore columns or exclude certain expectations, specify a threshold for creating value set expectations, or even specify semantic types for a given column.
You can find more information about [how to configure this profiler, including a list of the expectations that it uses, here.](https://docs.greatexpectations.io/en/latest/guides/how_to_guides/creating_and_editing_expectations/how_to_create_an_expectation_suite_with_the_user_configurable_profiler.html)

"""
        )
        self._add_scaffold_cell()
        self.add_footer()
        return self._notebook

    def render_to_disk(self, notebook_file_path: str) -> None:
        """
        Render a notebook to disk from an expectation suite.

        If batch_kwargs are passed they will override any found in suite
        citations.
        """
        self.render(self.batch_kwargs)
        self.write_notebook_to_disk(self._notebook, notebook_file_path)

    def _add_scaffold_cell(self) -> None:
        self.add_code_cell(
            """\
profiler = UserConfigurableProfiler(profile_dataset=batch,
    ignored_columns=ignored_columns,
    excluded_expectations=None,
    not_null_only=False,
    primary_or_compound_key=None,
    semantic_types_dict=None,
    table_expectations_only=False,
    value_set_threshold="MANY",
    )

suite = profiler.build_suite()"""
        )



import random
import uuid
from typing import Union

import great_expectations.exceptions as gx_exceptions
from great_expectations.data_context.cloud_constants import GXCloudRESTResource
from great_expectations.data_context.store.configuration_store import ConfigurationStore
from great_expectations.data_context.types.resource_identifiers import (
    ConfigurationIdentifier,  # noqa: TCH001
    GXCloudIdentifier,  # noqa: TCH001
)
from great_expectations.rule_based_profiler.config import RuleBasedProfilerConfig


class ProfilerStore(ConfigurationStore):
    """
    A ProfilerStore manages Profilers for the DataContext.
    """

    _configuration_class = RuleBasedProfilerConfig

    def serialization_self_check(self, pretty_print: bool) -> None:
        """
        Fufills the abstract method defined by the parent class.
        See `ConfigurationStore` for more details.
        """
        test_profiler_name = f"profiler_{''.join([random.choice(list('0123456789ABCDEF')) for _ in range(20)])}"
        test_profiler_configuration = RuleBasedProfilerConfig(
            name=test_profiler_name,
            config_version=1.0,
            rules={},
        )

        test_key: Union[GXCloudIdentifier, ConfigurationIdentifier]
        if self.cloud_mode:
            test_key = self.key_class(  # type: ignore[assignment,call-arg]
                resource_type=GXCloudRESTResource.PROFILER,
                ge_cloud_id=str(uuid.uuid4()),
            )
        else:
            test_key = self.key_class(configuration_key=test_profiler_name)  # type: ignore[assignment,call-arg]

        if pretty_print:
            print(f"Attempting to add a new test key {test_key} to Profiler store...")

        self.set(key=test_key, value=test_profiler_configuration)
        if pretty_print:
            print(f"\tTest key {test_key} successfully added to Profiler store.\n")
            print(
                f"Attempting to retrieve the test value associated with key {test_key} from Profiler store..."
            )

        test_value = self.get(key=test_key)
        if pretty_print:
            print(
                f"\tTest value successfully retrieved from Profiler store: {test_value}\n"
            )
            print(f"Cleaning up test key {test_key} and value from Profiler store...")

        test_value = self.remove_key(key=test_key)
        if pretty_print:
            print(
                f"\tTest key and value successfully removed from Profiler store: {test_value}\n"
            )

    def ge_cloud_response_json_to_object_dict(self, response_json: dict) -> dict:
        """
        This method takes full json response from GX cloud and outputs a dict appropriate for
        deserialization into a GX object
        """
        ge_cloud_profiler_id = response_json["data"]["id"]
        profiler_config_dict = response_json["data"]["attributes"]["profiler"]
        profiler_config_dict["id"] = ge_cloud_profiler_id

        return profiler_config_dict

    def add(self, key, value, **kwargs):
        try:
            return super().add(key=key, value=value, **kwargs)
        except gx_exceptions.StoreBackendError:
            raise gx_exceptions.ProfilerError(
                f"A Profiler named {value.name} already exists."
            )

    def update(self, key, value, **kwargs):
        try:
            return super().update(key=key, value=value, **kwargs)
        except gx_exceptions.StoreBackendError:
            raise gx_exceptions.ProfilerNotFoundError(
                f"Could not find an existing Profiler named {value.name}."
            )



from ruamel import yaml

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_gradual.py imports">
import great_expectations as gx

# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_gradual.py get_context">
context = gx.get_context()
# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_gradual.py datasource_yaml">
datasource_yaml = f"""
name: taxi_datasource
class_name: SimpleSqlalchemyDatasource
connection_string: <CONNECTION_STRING>

introspection:  # Each key in the "introspection" section is the name of an InferredAssetSqlDataConnector (key name "introspection" in "SimpleSqlalchemyDatasource" configuration is reserved).
    whole_table: {{}}  # Any alphanumeric key name is acceptable.
"""
# </snippet>

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your path directly in the yaml above.
data_dir_path = "data"
CONNECTION_STRING = f"sqlite:///{data_dir_path}/yellow_tripdata.db"

datasource_yaml = datasource_yaml.replace("<CONNECTION_STRING>", CONNECTION_STRING)

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_gradual.py test_yaml_config">
context.test_yaml_config(datasource_yaml)
# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_gradual.py buggy_datasource_yaml">
datasource_yaml = f"""  # buggy datasource_yaml configuration
name: mis_configured_datasource
class_name: SimpleSqlalchemyDatasource
connection_string: <CONNECTION_STRING>

introspecting:  # illegal top-level key name
    whole_table: {{}}
"""
# </snippet>

datasource_yaml = datasource_yaml.replace("<CONNECTION_STRING>", CONNECTION_STRING)

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_gradual.py test_yaml_config_2">
context.test_yaml_config(datasource_yaml)
# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_gradual.py datasource_yaml_introspection">
datasource_yaml = f"""
name: taxi_datasource
class_name: SimpleSqlalchemyDatasource
connection_string: <CONNECTION_STRING>

introspection:  # Each key in the "introspection" section is the name of an InferredAssetSqlDataConnector (key name "introspection" in "SimpleSqlalchemyDatasource" configuration is reserved).
    whole_table:
        include_schema_name: true
        introspection_directives:
            include_views: true
        skip_inapplicable_tables: true  # skip and continue upon encountering introspection errors
        excluded_tables:  # a list of tables to ignore when inferring data asset_names
            - main.yellow_tripdata_sample_2019_03  # format: schema_name.table_name
"""
# </snippet>

datasource_yaml = datasource_yaml.replace("<CONNECTION_STRING>", CONNECTION_STRING)

context.test_yaml_config(datasource_yaml)

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_gradual.py add_datasource">
context.add_datasource(**yaml.load(datasource_yaml))
# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_gradual.py get_available_data_asset_names">
available_data_asset_names = context.datasources[
    "taxi_datasource"
].get_available_data_asset_names(data_connector_names="whole_table")["whole_table"]
# </snippet>

assert len(available_data_asset_names) == 2

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_gradual.py datasource_yaml_tables">
datasource_yaml = f"""
name: taxi_datasource
class_name: SimpleSqlalchemyDatasource
connection_string: <CONNECTION_STRING>

tables:  # Each key in the "tables" section is a table_name (key name "tables" in "SimpleSqlalchemyDatasource" configuration is reserved).
    yellow_tripdata_sample_2019_01:  # Must match table name exactly.
        partitioners:  # Each key in the "partitioners" sub-section the name of a ConfiguredAssetSqlDataConnector (key name "partitioners" in "SimpleSqlalchemyDatasource" configuration is reserved).
            whole_table: {{}}
"""
# </snippet>

datasource_yaml = datasource_yaml.replace("<CONNECTION_STRING>", CONNECTION_STRING)

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_gradual.py test_yaml_config_3">
context.test_yaml_config(datasource_yaml)
# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_gradual.py datasource_yaml_tables_partitioners">
datasource_yaml = f"""
name: taxi_datasource
class_name: SimpleSqlalchemyDatasource
connection_string: <CONNECTION_STRING>

tables:  # Each key in the "tables" section is a table_name (key name "tables" in "SimpleSqlalchemyDatasource" configuration is reserved).
    yellow_tripdata_sample_2019_01:  # Must match table name exactly.
        partitioners:  # Each key in the "partitioners" sub-section the name of a ConfiguredAssetSqlDataConnector (key name "partitioners" in "SimpleSqlalchemyDatasource" configuration is reserved).
            whole_table:
                include_schema_name: true
                schema_name: main
                data_asset_name_prefix: taxi__
                data_asset_name_suffix: __asset
"""
# </snippet>

datasource_yaml = datasource_yaml.replace("<CONNECTION_STRING>", CONNECTION_STRING)

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_gradual.py test_yaml_config_4">
context.test_yaml_config(datasource_yaml)
# </snippet>

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/how_to_introspect_and_partition_your_data/sql_database/yaml_example_gradual.py add_datasource_2">
context.add_datasource(**yaml.load(datasource_yaml))
# </snippet>

available_data_asset_names = context.datasources[
    "taxi_datasource"
].get_available_data_asset_names(data_connector_names="whole_table")["whole_table"]
assert len(available_data_asset_names) == 1



from typing import Any, Dict, List, Optional
from unittest import mock
from unittest.mock import MagicMock

import pandas as pd
import pytest

import great_expectations.exceptions as gx_exceptions
from great_expectations.core.batch import BatchRequest
from great_expectations.data_context.data_context.cloud_data_context import (
    CloudDataContext,
)
from great_expectations.data_context.store.profiler_store import ProfilerStore
from great_expectations.data_context.types.resource_identifiers import (
    ConfigurationIdentifier,
    GXCloudIdentifier,
)
from great_expectations.exceptions.exceptions import InvalidConfigError
from great_expectations.rule_based_profiler import (
    BaseRuleBasedProfiler,
    RuleBasedProfiler,
    RuleBasedProfilerResult,
)
from great_expectations.rule_based_profiler.config import (
    DomainBuilderConfig,
    ParameterBuilderConfig,
    RuleBasedProfilerConfig,
)
from great_expectations.rule_based_profiler.domain_builder import TableDomainBuilder
from great_expectations.rule_based_profiler.expectation_configuration_builder import (
    DefaultExpectationConfigurationBuilder,
)
from great_expectations.rule_based_profiler.helpers.configuration_reconciliation import (
    ReconciliationDirectives,
    ReconciliationStrategy,
)
from great_expectations.rule_based_profiler.parameter_builder import (
    MetricMultiBatchParameterBuilder,
)
from great_expectations.rule_based_profiler.parameter_container import (
    ParameterContainer,
)
from great_expectations.rule_based_profiler.rule import Rule
from great_expectations.util import deep_filter_properties_iterable


@pytest.fixture()
def sample_rule_dict():
    return {
        "domain_builder": {
            "include_column_names": None,
            "module_name": "great_expectations.rule_based_profiler.domain_builder.column_domain_builder",
            "class_name": "ColumnDomainBuilder",
            "include_column_name_suffixes": ["_amount"],
        },
        "parameter_builders": [],
        "expectation_configuration_builders": [
            {
                "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
                "condition": None,
                "expectation_type": "expect_column_values_to_not_be_null",
                "meta": {},
                "column": "$domain.domain_kwargs.column",
                "class_name": "DefaultExpectationConfigurationBuilder",
            }
        ],
    }


@pytest.mark.unit
def test_reconcile_profiler_variables_no_overrides(
    profiler_with_placeholder_args,
    variables_multi_part_name_parameter_container,
):
    variables: Dict[str, Any] = {}
    effective_variables: Optional[
        ParameterContainer
    ] = profiler_with_placeholder_args.reconcile_profiler_variables(variables=variables)
    assert effective_variables == variables_multi_part_name_parameter_container


@pytest.mark.unit
def test_reconcile_profiler_variables_with_overrides(
    profiler_with_placeholder_args,
):
    variables: Dict[str, Any] = {
        "false_positive_threshold": 2.0e-2,
        "estimator": "bootstrap",
        "mostly": 8.0e-1,
    }
    effective_variables: Optional[
        ParameterContainer
    ] = profiler_with_placeholder_args.reconcile_profiler_variables(variables=variables)
    assert effective_variables.to_dict()["parameter_nodes"]["variables"][
        "variables"
    ] == {
        "false_positive_threshold": 2.0e-2,
        "estimator": "bootstrap",
        "mostly": 8.0e-1,
    }


@pytest.mark.unit
def test_reconcile_profiler_rules_no_overrides(
    profiler_with_placeholder_args,
):
    rules: Dict[str, Dict[str, Any]] = {}

    effective_rules: List[
        Rule
    ] = profiler_with_placeholder_args.reconcile_profiler_rules(rules=rules)
    assert effective_rules == profiler_with_placeholder_args.rules


@pytest.mark.unit
def test_reconcile_profiler_rules_new_rule_override(
    profiler_with_placeholder_args,
):
    rules: Dict[str, Dict[str, Any]] = {
        "rule_0": {
            "variables": {},
            "domain_builder": {
                "module_name": "great_expectations.rule_based_profiler.domain_builder",
                "class_name": "ColumnDomainBuilder",
            },
            "parameter_builders": [
                {
                    "class_name": "MetricMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder",
                    "name": "my_parameter",
                    "metric_name": "my_metric",
                },
                {
                    "class_name": "NumericMetricRangeMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder",
                    "name": "my_other_parameter",
                    "metric_name": "my_other_metric",
                    "quantile_statistic_interpolation_method": "auto",
                    "include_estimator_samples_histogram_in_details": False,
                },
            ],
            "expectation_configuration_builders": [
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder",
                    "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                    "column_A": "$domain.domain_kwargs.column_A",
                    "column_B": "$domain.domain_kwargs.column_B",
                    "my_one_arg": "$parameter.my_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_parameter_estimator": "$parameter.my_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder",
                    "expectation_type": "expect_column_min_to_be_between",
                    "column": "$domain.domain_kwargs.column",
                    "my_another_arg": "$parameter.my_other_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_other_parameter_estimator": "$parameter.my_other_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
            ],
        },
    }

    expected_rules: Dict[str, dict] = {
        "rule_0": {
            "variables": {},
            "domain_builder": {
                "module_name": "great_expectations.rule_based_profiler.domain_builder.column_domain_builder",
                "class_name": "ColumnDomainBuilder",
            },
            "parameter_builders": [
                {
                    "class_name": "MetricMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder.metric_multi_batch_parameter_builder",
                    "name": "my_parameter",
                    "metric_name": "my_metric",
                    "single_batch_mode": False,
                    "enforce_numeric_metric": False,
                    "replace_nan_with_zero": False,
                    "reduce_scalar_metric": True,
                },
                {
                    "class_name": "NumericMetricRangeMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder.numeric_metric_range_multi_batch_parameter_builder",
                    "name": "my_other_parameter",
                    "metric_name": "my_other_metric",
                    "estimator": "bootstrap",
                    "enforce_numeric_metric": True,
                    "replace_nan_with_zero": True,
                    "reduce_scalar_metric": True,
                    "false_positive_rate": 0.05,
                    "quantile_statistic_interpolation_method": "auto",
                    "quantile_bias_correction": False,
                    "include_estimator_samples_histogram_in_details": False,
                    "truncate_values": {},
                },
            ],
            "expectation_configuration_builders": [
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
                    "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                    "column_A": "$domain.domain_kwargs.column_A",
                    "column_B": "$domain.domain_kwargs.column_B",
                    "my_one_arg": "$parameter.my_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_parameter_estimator": "$parameter.my_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
                    "expectation_type": "expect_column_min_to_be_between",
                    "column": "$domain.domain_kwargs.column",
                    "my_another_arg": "$parameter.my_other_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_other_parameter_estimator": "$parameter.my_other_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
            ],
        },
        "rule_1": {
            "variables": {},
            "domain_builder": {
                "module_name": "great_expectations.rule_based_profiler.domain_builder.table_domain_builder",
                "class_name": "TableDomainBuilder",
            },
            "parameter_builders": [
                {
                    "class_name": "MetricMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder.metric_multi_batch_parameter_builder",
                    "name": "my_parameter",
                    "metric_name": "my_metric",
                    "single_batch_mode": False,
                    "enforce_numeric_metric": False,
                    "replace_nan_with_zero": False,
                    "reduce_scalar_metric": True,
                },
            ],
            "expectation_configuration_builders": [
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
                    "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                    "column_A": "$domain.domain_kwargs.column_A",
                    "column_B": "$domain.domain_kwargs.column_B",
                    "my_arg": "$parameter.my_parameter.value[0]",
                    "my_other_arg": "$parameter.my_parameter.value[1]",
                    "meta": {
                        "profiler_details": {
                            "my_parameter_estimator": "$parameter.my_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
            ],
        },
    }

    effective_rules: List[
        Rule
    ] = profiler_with_placeholder_args.reconcile_profiler_rules(rules=rules)

    rule: Rule
    effective_rule_configs_actual: Dict[str, dict] = {
        rule.name: rule.to_json_dict() for rule in effective_rules
    }
    deep_filter_properties_iterable(effective_rule_configs_actual, inplace=True)

    assert effective_rule_configs_actual == expected_rules


@pytest.mark.unit
def test_reconcile_profiler_rules_existing_rule_domain_builder_override(
    profiler_with_placeholder_args,
):
    rules: Dict[str, Dict[str, Any]] = {
        "rule_1": {
            "variables": {},
            "domain_builder": {
                "module_name": "great_expectations.rule_based_profiler.domain_builder",
                "class_name": "ColumnDomainBuilder",
                "include_column_name_suffixes": [
                    "_ts",
                ],
            },
        },
    }

    expected_rules: Dict[str, dict] = {
        "rule_1": {
            "variables": {},
            "domain_builder": {
                "module_name": "great_expectations.rule_based_profiler.domain_builder.column_domain_builder",
                "class_name": "ColumnDomainBuilder",
                "include_column_name_suffixes": [
                    "_ts",
                ],
            },
            "parameter_builders": [
                {
                    "class_name": "MetricMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder.metric_multi_batch_parameter_builder",
                    "name": "my_parameter",
                    "metric_name": "my_metric",
                    "single_batch_mode": False,
                    "enforce_numeric_metric": False,
                    "replace_nan_with_zero": False,
                    "reduce_scalar_metric": True,
                },
            ],
            "expectation_configuration_builders": [
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
                    "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                    "column_A": "$domain.domain_kwargs.column_A",
                    "column_B": "$domain.domain_kwargs.column_B",
                    "my_arg": "$parameter.my_parameter.value[0]",
                    "my_other_arg": "$parameter.my_parameter.value[1]",
                    "meta": {
                        "profiler_details": {
                            "my_parameter_estimator": "$parameter.my_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
            ],
        },
    }

    effective_rules: List[
        Rule
    ] = profiler_with_placeholder_args.reconcile_profiler_rules(rules=rules)

    rule: Rule
    effective_rule_configs_actual: Dict[str, dict] = {
        rule.name: rule.to_json_dict() for rule in effective_rules
    }
    deep_filter_properties_iterable(effective_rule_configs_actual, inplace=True)

    assert effective_rule_configs_actual == expected_rules


@pytest.mark.unit
def test_reconcile_profiler_rules_existing_rule_parameter_builder_overrides(
    profiler_with_placeholder_args,
):
    rules: Dict[str, Dict[str, Any]] = {
        "rule_1": {
            "parameter_builders": [
                {
                    "class_name": "MetricMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder",
                    "name": "my_parameter",
                    "metric_name": "my_special_metric",
                    "single_batch_mode": False,
                    "enforce_numeric_metric": True,
                    "replace_nan_with_zero": True,
                    "reduce_scalar_metric": True,
                },
                {
                    "class_name": "NumericMetricRangeMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder",
                    "name": "my_other_parameter",
                    "metric_name": "my_other_metric",
                    "enforce_numeric_metric": True,
                    "replace_nan_with_zero": False,
                    "reduce_scalar_metric": True,
                    "false_positive_rate": 0.025,
                    "quantile_statistic_interpolation_method": "auto",
                    "include_estimator_samples_histogram_in_details": False,
                },
            ],
        },
    }

    expected_rules: Dict[str, dict] = {
        "rule_1": {
            "variables": {},
            "domain_builder": {
                "module_name": "great_expectations.rule_based_profiler.domain_builder.table_domain_builder",
                "class_name": "TableDomainBuilder",
            },
            "parameter_builders": [
                {
                    "class_name": "MetricMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder.metric_multi_batch_parameter_builder",
                    "name": "my_parameter",
                    "metric_name": "my_special_metric",
                    "single_batch_mode": False,
                    "enforce_numeric_metric": True,
                    "replace_nan_with_zero": True,
                    "reduce_scalar_metric": True,
                },
                {
                    "class_name": "NumericMetricRangeMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder.numeric_metric_range_multi_batch_parameter_builder",
                    "name": "my_other_parameter",
                    "metric_name": "my_other_metric",
                    "estimator": "bootstrap",
                    "enforce_numeric_metric": True,
                    "replace_nan_with_zero": False,
                    "reduce_scalar_metric": True,
                    "false_positive_rate": 0.025,
                    "quantile_statistic_interpolation_method": "auto",
                    "quantile_bias_correction": False,
                    "include_estimator_samples_histogram_in_details": False,
                    "truncate_values": {},
                },
            ],
            "expectation_configuration_builders": [
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
                    "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                    "column_A": "$domain.domain_kwargs.column_A",
                    "column_B": "$domain.domain_kwargs.column_B",
                    "my_arg": "$parameter.my_parameter.value[0]",
                    "my_other_arg": "$parameter.my_parameter.value[1]",
                    "meta": {
                        "profiler_details": {
                            "my_parameter_estimator": "$parameter.my_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
            ],
        },
    }

    effective_rules: List[
        Rule
    ] = profiler_with_placeholder_args.reconcile_profiler_rules(rules=rules)

    rule: Rule
    effective_rule_configs_actual: Dict[str, dict] = {
        rule.name: rule.to_json_dict() for rule in effective_rules
    }
    deep_filter_properties_iterable(effective_rule_configs_actual, inplace=True)

    assert effective_rule_configs_actual == expected_rules


@pytest.mark.unit
def test_reconcile_profiler_rules_existing_rule_expectation_configuration_builder_overrides(
    profiler_with_placeholder_args,
):
    rules: Dict[str, Dict[str, Any]] = {
        "rule_1": {
            "expectation_configuration_builders": [
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder",
                    "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                    "column_A": "$domain.domain_kwargs.column_A",
                    "column_B": "$domain.domain_kwargs.column_B",
                    "my_one_arg": "$parameter.my_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_parameter_estimator": "$parameter.my_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder",
                    "expectation_type": "expect_column_min_to_be_between",
                    "column": "$domain.domain_kwargs.column",
                    "my_another_arg": "$parameter.my_other_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_other_parameter_estimator": "$parameter.my_other_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
            ],
        },
    }

    expected_rules: Dict[str, dict] = {
        "rule_1": {
            "variables": {},
            "domain_builder": {
                "module_name": "great_expectations.rule_based_profiler.domain_builder.table_domain_builder",
                "class_name": "TableDomainBuilder",
            },
            "parameter_builders": [
                {
                    "class_name": "MetricMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder.metric_multi_batch_parameter_builder",
                    "name": "my_parameter",
                    "metric_name": "my_metric",
                    "single_batch_mode": False,
                    "enforce_numeric_metric": False,
                    "replace_nan_with_zero": False,
                    "reduce_scalar_metric": True,
                },
            ],
            "expectation_configuration_builders": [
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
                    "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                    "column_A": "$domain.domain_kwargs.column_A",
                    "column_B": "$domain.domain_kwargs.column_B",
                    "my_one_arg": "$parameter.my_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_parameter_estimator": "$parameter.my_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
                    "expectation_type": "expect_column_min_to_be_between",
                    "column": "$domain.domain_kwargs.column",
                    "my_another_arg": "$parameter.my_other_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_other_parameter_estimator": "$parameter.my_other_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
            ],
        },
    }

    effective_rules: List[
        Rule
    ] = profiler_with_placeholder_args.reconcile_profiler_rules(rules=rules)

    rule: Rule
    effective_rule_configs_actual: Dict[str, dict] = {
        rule.name: rule.to_json_dict() for rule in effective_rules
    }
    deep_filter_properties_iterable(effective_rule_configs_actual, inplace=True)

    assert effective_rule_configs_actual == expected_rules


@pytest.mark.unit
def test_reconcile_profiler_rules_existing_rule_full_rule_override_nested_update(
    profiler_with_placeholder_args,
):
    rules: Dict[str, Dict[str, Any]] = {
        "rule_1": {
            "variables": {},
            "domain_builder": {
                "module_name": "great_expectations.rule_based_profiler.domain_builder",
                "class_name": "ColumnDomainBuilder",
            },
            "parameter_builders": [
                {
                    "class_name": "MetricMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder",
                    "name": "my_parameter",
                    "metric_name": "my_metric",
                },
                {
                    "class_name": "NumericMetricRangeMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder",
                    "name": "my_other_parameter",
                    "metric_name": "my_other_metric",
                    "quantile_statistic_interpolation_method": "auto",
                    "include_estimator_samples_histogram_in_details": False,
                },
            ],
            "expectation_configuration_builders": [
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder",
                    "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                    "column_A": "$domain.domain_kwargs.column_A",
                    "column_B": "$domain.domain_kwargs.column_B",
                    "my_one_arg": "$parameter.my_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_parameter_estimator": "$parameter.my_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder",
                    "expectation_type": "expect_column_min_to_be_between",
                    "column": "$domain.domain_kwargs.column",
                    "my_another_arg": "$parameter.my_other_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_other_parameter_estimator": "$parameter.my_other_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
            ],
        },
    }

    expected_rules: Dict[str, dict] = {
        "rule_1": {
            "variables": {},
            "domain_builder": {
                "module_name": "great_expectations.rule_based_profiler.domain_builder.column_domain_builder",
                "class_name": "ColumnDomainBuilder",
            },
            "parameter_builders": [
                {
                    "class_name": "MetricMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder.metric_multi_batch_parameter_builder",
                    "name": "my_parameter",
                    "metric_name": "my_metric",
                    "single_batch_mode": False,
                    "enforce_numeric_metric": False,
                    "replace_nan_with_zero": False,
                    "reduce_scalar_metric": True,
                },
                {
                    "class_name": "NumericMetricRangeMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder.numeric_metric_range_multi_batch_parameter_builder",
                    "name": "my_other_parameter",
                    "metric_name": "my_other_metric",
                    "estimator": "bootstrap",
                    "enforce_numeric_metric": True,
                    "replace_nan_with_zero": True,
                    "reduce_scalar_metric": True,
                    "false_positive_rate": 0.05,
                    "quantile_statistic_interpolation_method": "auto",
                    "quantile_bias_correction": False,
                    "include_estimator_samples_histogram_in_details": False,
                    "truncate_values": {},
                },
            ],
            "expectation_configuration_builders": [
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
                    "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                    "column_A": "$domain.domain_kwargs.column_A",
                    "column_B": "$domain.domain_kwargs.column_B",
                    "my_arg": "$parameter.my_parameter.value[0]",
                    "my_other_arg": "$parameter.my_parameter.value[1]",
                    "my_one_arg": "$parameter.my_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_parameter_estimator": "$parameter.my_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
                    "expectation_type": "expect_column_min_to_be_between",
                    "column": "$domain.domain_kwargs.column",
                    "my_another_arg": "$parameter.my_other_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_other_parameter_estimator": "$parameter.my_other_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
            ],
        },
    }

    effective_rules: List[
        Rule
    ] = profiler_with_placeholder_args.reconcile_profiler_rules(
        rules=rules,
        reconciliation_directives=ReconciliationDirectives(
            domain_builder=ReconciliationStrategy.UPDATE,
            parameter_builder=ReconciliationStrategy.UPDATE,
            expectation_configuration_builder=ReconciliationStrategy.NESTED_UPDATE,
        ),
    )

    rule: Rule
    effective_rule_configs_actual: Dict[str, dict] = {
        rule.name: rule.to_json_dict() for rule in effective_rules
    }
    deep_filter_properties_iterable(effective_rule_configs_actual, inplace=True)

    assert effective_rule_configs_actual == expected_rules


@pytest.mark.unit
def test_reconcile_profiler_rules_existing_rule_full_rule_override_replace(
    profiler_with_placeholder_args,
):
    rules: Dict[str, Dict[str, Any]] = {
        "rule_1": {
            "variables": {},
            "domain_builder": {
                "module_name": "great_expectations.rule_based_profiler.domain_builder",
                "class_name": "ColumnDomainBuilder",
            },
            "parameter_builders": [
                {
                    "class_name": "NumericMetricRangeMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder",
                    "name": "my_other_parameter",
                    "metric_name": "my_other_metric",
                    "quantile_statistic_interpolation_method": "auto",
                    "include_estimator_samples_histogram_in_details": False,
                },
            ],
            "expectation_configuration_builders": [
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder",
                    "expectation_type": "expect_column_min_to_be_between",
                    "column": "$domain.domain_kwargs.column",
                    "my_another_arg": "$parameter.my_other_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_other_parameter_estimator": "$parameter.my_other_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
            ],
        },
    }

    expected_rules: Dict[str, Dict] = {
        "rule_1": {
            "variables": {},
            "domain_builder": {
                "module_name": "great_expectations.rule_based_profiler.domain_builder.column_domain_builder",
                "class_name": "ColumnDomainBuilder",
            },
            "parameter_builders": [
                {
                    "class_name": "NumericMetricRangeMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder.numeric_metric_range_multi_batch_parameter_builder",
                    "name": "my_other_parameter",
                    "metric_name": "my_other_metric",
                    "estimator": "bootstrap",
                    "enforce_numeric_metric": True,
                    "replace_nan_with_zero": True,
                    "reduce_scalar_metric": True,
                    "false_positive_rate": 0.05,
                    "quantile_statistic_interpolation_method": "auto",
                    "quantile_bias_correction": False,
                    "include_estimator_samples_histogram_in_details": False,
                    "truncate_values": {},
                },
            ],
            "expectation_configuration_builders": [
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
                    "expectation_type": "expect_column_min_to_be_between",
                    "column": "$domain.domain_kwargs.column",
                    "my_another_arg": "$parameter.my_other_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_other_parameter_estimator": "$parameter.my_other_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
            ],
        },
    }

    effective_rules: List[
        Rule
    ] = profiler_with_placeholder_args.reconcile_profiler_rules(
        rules=rules,
        reconciliation_directives=ReconciliationDirectives(
            domain_builder=ReconciliationStrategy.UPDATE,
            parameter_builder=ReconciliationStrategy.REPLACE,
            expectation_configuration_builder=ReconciliationStrategy.REPLACE,
        ),
    )

    rule: Rule
    effective_rule_configs_actual: Dict[str, dict] = {
        rule.name: rule.to_json_dict() for rule in effective_rules
    }
    deep_filter_properties_iterable(effective_rule_configs_actual, inplace=True)

    assert effective_rule_configs_actual == expected_rules


@pytest.mark.unit
def test_reconcile_profiler_rules_existing_rule_full_rule_override_update(
    profiler_with_placeholder_args,
):
    rules: Dict[str, Dict[str, Any]] = {
        "rule_1": {
            "variables": {},
            "domain_builder": {
                "module_name": "great_expectations.rule_based_profiler.domain_builder",
                "class_name": "ColumnDomainBuilder",
            },
            "parameter_builders": [
                {
                    "class_name": "MetricMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder",
                    "name": "my_parameter",
                    "metric_name": "my_metric",
                },
                {
                    "class_name": "NumericMetricRangeMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder",
                    "name": "my_other_parameter",
                    "metric_name": "my_other_metric",
                    "quantile_statistic_interpolation_method": "auto",
                    "include_estimator_samples_histogram_in_details": False,
                },
            ],
            "expectation_configuration_builders": [
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder",
                    "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                    "column_A": "$domain.domain_kwargs.column_A",
                    "column_B": "$domain.domain_kwargs.column_B",
                    "my_one_arg": "$parameter.my_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_parameter_estimator": "$parameter.my_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder",
                    "expectation_type": "expect_column_min_to_be_between",
                    "column": "$domain.domain_kwargs.column",
                    "my_another_arg": "$parameter.my_other_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_other_parameter_estimator": "$parameter.my_other_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
            ],
        },
    }

    expected_rules: Dict[str, dict] = {
        "rule_1": {
            "variables": {},
            "domain_builder": {
                "module_name": "great_expectations.rule_based_profiler.domain_builder.column_domain_builder",
                "class_name": "ColumnDomainBuilder",
            },
            "parameter_builders": [
                {
                    "class_name": "MetricMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder.metric_multi_batch_parameter_builder",
                    "name": "my_parameter",
                    "metric_name": "my_metric",
                    "single_batch_mode": False,
                    "enforce_numeric_metric": False,
                    "replace_nan_with_zero": False,
                    "reduce_scalar_metric": True,
                },
                {
                    "class_name": "NumericMetricRangeMultiBatchParameterBuilder",
                    "module_name": "great_expectations.rule_based_profiler.parameter_builder.numeric_metric_range_multi_batch_parameter_builder",
                    "name": "my_other_parameter",
                    "metric_name": "my_other_metric",
                    "estimator": "bootstrap",
                    "enforce_numeric_metric": True,
                    "replace_nan_with_zero": True,
                    "reduce_scalar_metric": True,
                    "false_positive_rate": 0.05,
                    "quantile_statistic_interpolation_method": "auto",
                    "quantile_bias_correction": False,
                    "include_estimator_samples_histogram_in_details": False,
                    "truncate_values": {},
                },
            ],
            "expectation_configuration_builders": [
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
                    "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                    "column_A": "$domain.domain_kwargs.column_A",
                    "column_B": "$domain.domain_kwargs.column_B",
                    "my_one_arg": "$parameter.my_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_parameter_estimator": "$parameter.my_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
                    "expectation_type": "expect_column_min_to_be_between",
                    "column": "$domain.domain_kwargs.column",
                    "my_another_arg": "$parameter.my_other_parameter.value[0]",
                    "meta": {
                        "profiler_details": {
                            "my_other_parameter_estimator": "$parameter.my_other_parameter.details",
                            "note": "Important remarks about estimation algorithm.",
                        },
                    },
                },
            ],
        },
    }

    effective_rules: List[
        Rule
    ] = profiler_with_placeholder_args.reconcile_profiler_rules(rules=rules)

    rule: Rule
    effective_rule_configs_actual: Dict[str, dict] = {
        rule.name: rule.to_json_dict() for rule in effective_rules
    }
    deep_filter_properties_iterable(effective_rule_configs_actual, inplace=True)

    assert effective_rule_configs_actual == expected_rules


@mock.patch("great_expectations.rule_based_profiler.RuleBasedProfiler.run")
@mock.patch("great_expectations.data_context.data_context.AbstractDataContext")
@pytest.mark.unit
def test_run_profiler_without_dynamic_args(
    mock_data_context: mock.MagicMock,
    mock_profiler_run: mock.MagicMock,
    populated_profiler_store: ProfilerStore,
    profiler_name: str,
):
    RuleBasedProfiler.run_profiler(
        data_context=mock_data_context,
        profiler_store=populated_profiler_store,
        name=profiler_name,
    )

    assert mock_profiler_run.called
    assert mock_profiler_run.call_args == mock.call(
        variables=None,
        rules=None,
        batch_list=None,
        batch_request=None,
        runtime_configuration=None,
        reconciliation_directives=ReconciliationDirectives(
            variables=ReconciliationStrategy.UPDATE,
            domain_builder=ReconciliationStrategy.UPDATE,
            parameter_builder=ReconciliationStrategy.UPDATE,
            expectation_configuration_builder=ReconciliationStrategy.UPDATE,
        ),
        variables_directives_list=None,
        domain_type_directives_list=None,
        comment=None,
    )


@mock.patch("great_expectations.rule_based_profiler.RuleBasedProfiler.run")
@mock.patch("great_expectations.data_context.data_context.AbstractDataContext")
@pytest.mark.unit
def test_run_profiler_with_dynamic_args(
    mock_data_context: mock.MagicMock,
    mock_profiler_run: mock.MagicMock,
    populated_profiler_store: ProfilerStore,
    profiler_name: str,
):
    # Dynamic arguments used to override the profiler's attributes
    variables = {"foo": "bar"}
    rules = {"baz": "qux"}

    # noinspection PyUnusedLocal
    rule_based_profiler_result: RuleBasedProfilerResult = (
        RuleBasedProfiler.run_profiler(
            data_context=mock_data_context,
            profiler_store=populated_profiler_store,
            name=profiler_name,
            variables=variables,
            rules=rules,
        )
    )

    assert mock_profiler_run.called
    assert mock_profiler_run.call_args == mock.call(
        variables=variables,
        rules=rules,
        batch_list=None,
        batch_request=None,
        runtime_configuration=None,
        reconciliation_directives=ReconciliationDirectives(
            variables=ReconciliationStrategy.UPDATE,
            domain_builder=ReconciliationStrategy.UPDATE,
            parameter_builder=ReconciliationStrategy.UPDATE,
            expectation_configuration_builder=ReconciliationStrategy.UPDATE,
        ),
        variables_directives_list=None,
        domain_type_directives_list=None,
        comment=None,
    )


@mock.patch("great_expectations.rule_based_profiler.RuleBasedProfiler.run")
@mock.patch("great_expectations.data_context.data_context.AbstractDataContext")
@pytest.mark.unit
def test_run_profiler_on_data_creates_suite_with_dict_arg(
    mock_data_context: mock.MagicMock,
    mock_rule_based_profiler_run: mock.MagicMock,
    populated_profiler_store: ProfilerStore,
    profiler_name: str,
):
    batch_request: Dict[str, str] = {
        "datasource_name": "my_datasource",
        "data_connector_name": "my_data_connector",
        "data_asset_name": "my_data_asset",
    }

    RuleBasedProfiler.run_profiler_on_data(
        data_context=mock_data_context,
        profiler_store=populated_profiler_store,
        name=profiler_name,
        batch_request=batch_request,
    )

    assert mock_rule_based_profiler_run.called

    resulting_batch_request = mock_rule_based_profiler_run.call_args[1]["batch_request"]
    assert resulting_batch_request == batch_request


@mock.patch("great_expectations.rule_based_profiler.RuleBasedProfiler.run")
@mock.patch("great_expectations.data_context.data_context.AbstractDataContext")
@pytest.mark.unit
def test_run_profiler_on_data_creates_suite_with_batch_request_arg(
    mock_data_context: mock.MagicMock,
    mock_rule_based_profiler_run: mock.MagicMock,
    populated_profiler_store: ProfilerStore,
    profiler_name: str,
):
    batch_request: BatchRequest = BatchRequest(
        datasource_name="my_datasource",
        data_connector_name="my_data_connector",
        data_asset_name="my_data_asset",
    )

    RuleBasedProfiler.run_profiler_on_data(
        data_context=mock_data_context,
        profiler_store=populated_profiler_store,
        name=profiler_name,
        batch_request=batch_request,
    )

    assert mock_rule_based_profiler_run.called

    resulting_batch_request: dict = mock_rule_based_profiler_run.call_args[1][
        "batch_request"
    ].to_json_dict()
    deep_filter_properties_iterable(resulting_batch_request, inplace=True)
    expected_batch_request: dict = batch_request.to_json_dict()
    deep_filter_properties_iterable(expected_batch_request, inplace=True)
    assert resulting_batch_request == expected_batch_request


@mock.patch("great_expectations.data_context.data_context.AbstractDataContext")
@pytest.mark.unit
def test_get_profiler_with_too_many_args_raises_error(
    mock_data_context: mock.MagicMock,
    populated_profiler_store: ProfilerStore,
):
    with pytest.raises(AssertionError) as e:
        RuleBasedProfiler.get_profiler(
            data_context=mock_data_context,
            profiler_store=populated_profiler_store,
            name="my_profiler",
            id="my_ge_cloud_id",
        )

    assert "either name or id" in str(e.value)


@mock.patch("great_expectations.data_context.data_context.AbstractDataContext")
@pytest.mark.unit
def test_serialize_profiler_config(
    mock_data_context: mock.MagicMock,
    profiler_config_with_placeholder_args: RuleBasedProfilerConfig,
):
    profiler = BaseRuleBasedProfiler(
        profiler_config=profiler_config_with_placeholder_args,
        data_context=mock_data_context,
    )
    assert profiler.config == profiler_config_with_placeholder_args
    assert len(profiler.rules) == 1
    assert isinstance(profiler.rules[0].domain_builder, TableDomainBuilder)
    assert DomainBuilderConfig(
        **profiler.rules[0].domain_builder.to_json_dict()
    ).to_json_dict() == {
        "module_name": "great_expectations.rule_based_profiler.domain_builder.table_domain_builder",
        "class_name": "TableDomainBuilder",
    }
    assert isinstance(
        profiler.rules[0].parameter_builders[0], MetricMultiBatchParameterBuilder
    )
    assert ParameterBuilderConfig(
        **profiler.rules[0].parameter_builders[0].to_json_dict()
    ).to_json_dict() == {
        "module_name": "great_expectations.rule_based_profiler.parameter_builder.metric_multi_batch_parameter_builder",
        "class_name": "MetricMultiBatchParameterBuilder",
        "name": "my_parameter",
        "metric_name": "my_metric",
        "metric_domain_kwargs": None,
        "metric_value_kwargs": None,
        "single_batch_mode": False,
        "enforce_numeric_metric": False,
        "replace_nan_with_zero": False,
        "reduce_scalar_metric": True,
        "evaluation_parameter_builder_configs": None,
    }
    assert isinstance(
        profiler.rules[0].expectation_configuration_builders[0],
        DefaultExpectationConfigurationBuilder,
    )
    assert DefaultExpectationConfigurationBuilder(
        **profiler.rules[0].expectation_configuration_builders[0].to_json_dict()
    ).to_json_dict() == {
        "module_name": "great_expectations.rule_based_profiler.expectation_configuration_builder.default_expectation_configuration_builder",
        "class_name": "DefaultExpectationConfigurationBuilder",
        "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
        "validation_parameter_builder_configs": None,
        "column_A": "$domain.domain_kwargs.column_A",
        "column_B": "$domain.domain_kwargs.column_B",
        "condition": None,
        "my_arg": "$parameter.my_parameter.value[0]",
        "my_other_arg": "$parameter.my_parameter.value[1]",
        "meta": {
            "profiler_details": {
                "my_parameter_estimator": "$parameter.my_parameter.details",
                "note": "Important remarks about estimation algorithm.",
            },
        },
    }


@mock.patch("great_expectations.data_context.data_context.AbstractDataContext")
@pytest.mark.unit
def test_add_profiler(
    mock_data_context: mock.MagicMock,
    profiler_key: ConfigurationIdentifier,
    profiler_config_with_placeholder_args: RuleBasedProfilerConfig,
):
    mock_data_context.cloud_mode = False

    profiler_args = profiler_config_with_placeholder_args.to_dict()
    for attr in ("class_name", "module_name"):
        profiler_args.pop(attr)

    profiler: RuleBasedProfiler = RuleBasedProfiler.add_profiler(
        data_context=mock_data_context,
        profiler_store=mock_data_context.profiler_store,
        **profiler_args,
    )

    assert isinstance(profiler, RuleBasedProfiler)
    assert profiler.name == profiler_config_with_placeholder_args.name
    mock_data_context.profiler_store.add.asset_called_once()


@pytest.mark.cloud
@pytest.mark.unit
def test_add_profiler_ge_cloud_mode(
    ge_cloud_profiler_id: str,
    ge_cloud_profiler_key: GXCloudIdentifier,
    profiler_config_with_placeholder_args: RuleBasedProfilerConfig,
):
    profiler_args = profiler_config_with_placeholder_args.to_dict()
    for attr in ("class_name", "module_name"):
        profiler_args.pop(attr)

    with mock.patch(
        "great_expectations.data_context.data_context.CloudDataContext",
        spec=CloudDataContext,
    ) as mock_data_context:
        profiler: RuleBasedProfiler = RuleBasedProfiler.add_profiler(
            data_context=mock_data_context,
            profiler_store=mock_data_context.profiler_store,
            **profiler_args,
        )

    assert isinstance(profiler, RuleBasedProfiler)
    assert profiler.name == profiler_config_with_placeholder_args.name
    mock_data_context.profiler_store.add.assert_called_once()


@mock.patch("great_expectations.data_context.data_context.AbstractDataContext")
@pytest.mark.unit
def test_add_profiler_with_batch_request_containing_batch_data_raises_error(
    mock_data_context: mock.MagicMock,
):
    name = "my_profiler_config"
    config_version = 1.0
    rules = {
        "rule_1": {
            "domain_builder": {
                "class_name": "TableDomainBuilder",
                "batch_request": {
                    "runtime_parameters": {
                        "batch_data": pd.DataFrame()  # Cannot be serialized in store
                    }
                },
            },
            "parameter_builders": [
                {
                    "class_name": "MetricMultiBatchParameterBuilder",
                    "name": "my_parameter",
                    "metric_name": "my_metric",
                },
            ],
            "expectation_configuration_builders": [
                {
                    "class_name": "DefaultExpectationConfigurationBuilder",
                    "expectation_type": "expect_column_pair_values_A_to_be_greater_than_B",
                },
            ],
        }
    }

    with pytest.raises(InvalidConfigError) as e:
        RuleBasedProfiler.add_profiler(
            data_context=mock_data_context,
            profiler_store=mock_data_context.profiler_store,
            name=name,
            config_version=config_version,
            rules=rules,
        )

    assert "batch_data found in batch_request" in str(e.value)


@mock.patch("great_expectations.data_context.data_context.AbstractDataContext")
@pytest.mark.unit
def test_get_profiler(
    mock_data_context: mock.MagicMock,
    populated_profiler_store: ProfilerStore,
    profiler_config_with_placeholder_args: RuleBasedProfilerConfig,
):
    with mock.patch(
        "great_expectations.data_context.store.profiler_store.ProfilerStore.get",
        return_value=profiler_config_with_placeholder_args,
    ):
        profiler: RuleBasedProfiler = RuleBasedProfiler.get_profiler(
            data_context=mock_data_context,
            profiler_store=populated_profiler_store,
            name="my_profiler",
            id=None,
        )

    assert isinstance(profiler, RuleBasedProfiler)


@mock.patch("great_expectations.data_context.data_context.AbstractDataContext")
@pytest.mark.unit
def test_get_profiler_non_existent_profiler_raises_error(
    mock_data_context: mock.MagicMock, empty_profiler_store: ProfilerStore
):
    with pytest.raises(gx_exceptions.ProfilerNotFoundError) as e:
        RuleBasedProfiler.get_profiler(
            data_context=mock_data_context,
            profiler_store=empty_profiler_store,
            name="my_profiler",
            id=None,
        )

    assert "Non-existent Profiler" in str(e.value)


@pytest.mark.unit
def test_delete_profiler(
    populated_profiler_store: ProfilerStore,
):
    with mock.patch(
        "great_expectations.data_context.store.profiler_store.ProfilerStore.remove_key",
    ) as mock_remove_key:
        RuleBasedProfiler.delete_profiler(
            profiler_store=populated_profiler_store,
            name="my_profiler",
            id=None,
        )

    assert mock_remove_key.call_count == 1
    assert mock_remove_key.call_args == mock.call(
        key=ConfigurationIdentifier("my_profiler")
    )


@pytest.mark.unit
def test_delete_profiler_with_too_many_args_raises_error(
    populated_profiler_store: ProfilerStore,
):
    with pytest.raises(AssertionError) as e:
        RuleBasedProfiler.delete_profiler(
            profiler_store=populated_profiler_store,
            name="my_profiler",
            id="my_ge_cloud_id",
        )

    assert "either name or id" in str(e.value)


@pytest.mark.unit
def test_delete_profiler_non_existent_profiler_raises_error(
    populated_profiler_store: ProfilerStore,
):
    with pytest.raises(gx_exceptions.ProfilerNotFoundError) as e:
        RuleBasedProfiler.delete_profiler(
            profiler_store=populated_profiler_store,
            name="my_non_existent_profiler",
            id=None,
        )

    assert "Non-existent Profiler" in str(e.value)


@mock.patch("great_expectations.data_context.store.ProfilerStore")
@pytest.mark.unit
def test_list_profilers(mock_profiler_store: mock.MagicMock):
    store = mock_profiler_store()
    keys = ["a", "b", "c"]
    store.list_keys.return_value = [ConfigurationIdentifier(char) for char in keys]
    res = RuleBasedProfiler.list_profilers(store, ge_cloud_mode=False)

    assert res == keys
    assert store.list_keys.called


@mock.patch("great_expectations.data_context.store.ProfilerStore")
@pytest.mark.cloud
@pytest.mark.unit
def test_list_profilers_in_cloud_mode(mock_profiler_store: mock.MagicMock):
    store = mock_profiler_store()
    keys = ["a", "b", "c"]
    store.list_keys.return_value = keys
    res = RuleBasedProfiler.list_profilers(store, ge_cloud_mode=True)

    assert res == keys
    assert store.list_keys.called


@mock.patch("great_expectations.data_context.data_context.AbstractDataContext")
@mock.patch("great_expectations.rule_based_profiler.domain_builder.ColumnDomainBuilder")
@mock.patch(
    "great_expectations.rule_based_profiler.expectation_configuration_builder.DefaultExpectationConfigurationBuilder"
)
@pytest.mark.unit
def test_add_single_rule(
    mock_expectation_configuration_builder: mock.MagicMock,
    mock_domain_builder: mock.MagicMock,
    mock_data_context: mock.MagicMock,
    sample_rule_dict: dict,
):
    profiler: RuleBasedProfiler = RuleBasedProfiler(
        name="my_rbp",
        config_version=1.0,
        data_context=mock_data_context,
    )
    first_rule = Rule(
        name="first_rule",
        variables=None,
        domain_builder=mock_domain_builder,
        expectation_configuration_builders=mock_expectation_configuration_builder,
    )
    first_rule.to_json_dict = MagicMock(return_value=sample_rule_dict)
    profiler.add_rule(rule=first_rule)
    assert len(profiler.rules) == 1

    duplicate_of_first_rule = Rule(
        name="first_rule",
        variables=None,
        domain_builder=mock_domain_builder,
        expectation_configuration_builders=mock_expectation_configuration_builder,
    )
    duplicate_of_first_rule.to_json_dict = MagicMock(return_value=sample_rule_dict)
    profiler.add_rule(rule=duplicate_of_first_rule)
    assert len(profiler.rules) == 1


@mock.patch("great_expectations.data_context.data_context.AbstractDataContext")
@mock.patch("great_expectations.rule_based_profiler.domain_builder.ColumnDomainBuilder")
@mock.patch(
    "great_expectations.rule_based_profiler.expectation_configuration_builder.DefaultExpectationConfigurationBuilder"
)
@pytest.mark.unit
def test_add_rule_overwrite_first_rule(
    mock_expectation_configuration_builder: mock.MagicMock,
    mock_domain_builder: mock.MagicMock,
    mock_data_context: mock.MagicMock,
    sample_rule_dict: dict,
):

    profiler: RuleBasedProfiler = RuleBasedProfiler(
        name="my_rbp",
        config_version=1.0,
        data_context=mock_data_context,
    )
    first_rule = Rule(
        name="first_rule",
        variables=None,
        domain_builder=mock_domain_builder,
        expectation_configuration_builders=mock_expectation_configuration_builder,
    )
    first_rule.to_json_dict = MagicMock(return_value=sample_rule_dict)
    profiler.add_rule(rule=first_rule)
    assert len(profiler.rules) == 1


@mock.patch("great_expectations.data_context.data_context.AbstractDataContext")
@mock.patch("great_expectations.rule_based_profiler.domain_builder.ColumnDomainBuilder")
@mock.patch(
    "great_expectations.rule_based_profiler.expectation_configuration_builder.DefaultExpectationConfigurationBuilder"
)
@pytest.mark.unit
def test_add_rule_add_second_rule(
    mock_expectation_configuration_builder: mock.MagicMock,
    mock_domain_builder: mock.MagicMock,
    mock_data_context: mock.MagicMock,
    sample_rule_dict: dict,
):
    profiler: RuleBasedProfiler = RuleBasedProfiler(
        name="my_rbp",
        config_version=1.0,
        data_context=mock_data_context,
    )
    first_rule = Rule(
        name="first_rule",
        variables=None,
        domain_builder=mock_domain_builder,
        expectation_configuration_builders=mock_expectation_configuration_builder,
    )
    first_rule.to_json_dict = MagicMock(return_value=sample_rule_dict)
    profiler.add_rule(rule=first_rule)
    assert len(profiler.rules) == 1

    second_rule = Rule(
        name="second_rule",
        variables=None,
        domain_builder=mock_domain_builder,
        expectation_configuration_builders=mock_expectation_configuration_builder,
    )
    second_rule.to_json_dict = MagicMock(return_value=sample_rule_dict)
    profiler.add_rule(rule=second_rule)
    assert len(profiler.rules) == 2


@mock.patch("great_expectations.data_context.data_context.AbstractDataContext")
@pytest.mark.unit
def test_add_rule_bad_rule(
    mock_data_context: mock.MagicMock,
):
    profiler: RuleBasedProfiler = RuleBasedProfiler(
        name="my_rbp",
        config_version=1.0,
        data_context=mock_data_context,
    )
    not_a_rule: dict = {
        "name": "first_rule",
        "domain_builder": "domain_builder",
        "expectation_configuration_builder": "expectation_configuration_builder",
    }
    with pytest.raises(AttributeError) as e:
        # noinspection PyTypeChecker
        profiler.add_rule(rule=not_a_rule)
    assert "'dict' object has no attribute 'name'" in str(e.value)



def test_spark_null_filters(spark_session):
    import pandas as pd
    import pyspark
    import pyspark.sql.functions as F

    df = spark_session.createDataFrame(
        pd.DataFrame({"a": [1.0, 2, 3, 4]}),
        schema=pyspark.sql.types.StructType(
            [pyspark.sql.types.StructField("a", pyspark.sql.types.FloatType())]
        ),
    )
    assert df.agg(F.max(F.col("a"))).collect()[0][0] == 4

    df = spark_session.createDataFrame(
        pd.DataFrame({"a": [1, 2, 3, 4]}),
        schema=pyspark.sql.types.StructType(
            [pyspark.sql.types.StructField("a", pyspark.sql.types.IntegerType())]
        ),
    )
    assert df.agg(F.max(F.col("a"))).collect()[0][0] == 4

    df = spark_session.createDataFrame(
        pd.DataFrame({"a": [1.0, 2, 3, None, None, 4]}),
        schema=pyspark.sql.types.StructType(
            [
                pyspark.sql.types.StructField(
                    "a", pyspark.sql.types.FloatType(), nullable=True
                )
            ]
        ),
    )
    assert (
        df.agg(
            F.max(F.when(~F.isnull(F.col("a")) & ~F.isnan(F.col("a")), F.col("a")))
        ).collect()[0][0]
        == 4
    )

    df = spark_session.createDataFrame(
        pd.DataFrame({"a": [1.0, 2, 3, None, None, 4]}),
        schema=pyspark.sql.types.StructType(
            [pyspark.sql.types.StructField("a", pyspark.sql.types.FloatType())]
        ),
    )
    assert (
        df.agg(
            F.max(
                F.when(
                    ~F.isnull(F.col("a")) & ~F.isnan(F.col("a")), F.col("a")
                ).otherwise(0)
            )
        ).collect()[0][0]
        == 4
    )


def test_sa_null_filters(sa):
    import pandas as pd

    eng = sa.create_engine("sqlite://")
    # Demonstrate that spark's max aggregate function can tolerate null values
    df = pd.DataFrame({"a": [1, 2, 3, None, None, 4]})
    df.to_sql(name="test", con=eng, index=False)

    assert eng.execute("SELECT MAX(a) FROM test;").fetchone()[0] == 4



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Dict, List, Optional, Set, Union

import numpy as np

from great_expectations.core.domain import Domain  # noqa: TCH001
from great_expectations.rule_based_profiler.config import (
    ParameterBuilderConfig,  # noqa: TCH001
)
from great_expectations.rule_based_profiler.helpers.util import (
    get_parameter_value_and_validate_return_type,
)
from great_expectations.rule_based_profiler.parameter_builder import (
    MetricMultiBatchParameterBuilder,
)
from great_expectations.rule_based_profiler.parameter_container import (
    FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY,
    FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY,
    ParameterContainer,
    ParameterNode,
)
from great_expectations.types.attributes import Attributes
from great_expectations.validator.computed_metric import MetricValue  # noqa: TCH001

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )
    from great_expectations.rule_based_profiler.metric_computation_result import (
        MetricValues,
    )


class MeanTableColumnsSetMatchMultiBatchParameterBuilder(
    MetricMultiBatchParameterBuilder
):
    """
    Compute mean match ratio (as a fraction) of "table.columns" metric across every Batch of data given.

    Step-1: Compute "table.columns" metric value for each Batch object.
    Step-2: Compute set union operation of column lists from Step-1 over all Batch objects (gives maximum column set).
    Step-3: Assign match scores: if column set of a Batch equals overall (maximum) column set, give it 1; 0 otherwise.
    Step-4: Compute mean value of match scores as "success_ratio" (divide sum of scores by number of Batch objects).
    """

    exclude_field_names: ClassVar[
        Set[str]
    ] = MetricMultiBatchParameterBuilder.exclude_field_names | {
        "metric_name",
        "single_batch_mode",
        "enforce_numeric_metric",
        "replace_nan_with_zero",
        "reduce_scalar_metric",
    }

    def __init__(
        self,
        name: str,
        metric_domain_kwargs: Optional[Union[str, dict]] = None,
        metric_value_kwargs: Optional[Union[str, dict]] = None,
        evaluation_parameter_builder_configs: Optional[
            List[ParameterBuilderConfig]
        ] = None,
        data_context: Optional[AbstractDataContext] = None,
    ) -> None:
        """
        Args:
            name: the name of this parameter -- this is user-specified parameter name (from configuration);
            it is not the fully-qualified parameter name; a fully-qualified parameter name must start with "$parameter."
            and may contain one or more subsequent parts (e.g., "$parameter.<my_param_from_config>.<metric_name>").
            metric_domain_kwargs: used in MetricConfiguration
            metric_value_kwargs: used in MetricConfiguration
            evaluation_parameter_builder_configs: ParameterBuilder configurations, executing and making whose respective
            ParameterBuilder objects' outputs available (as fully-qualified parameter names) is pre-requisite.
            These "ParameterBuilder" configurations help build parameters needed for this "ParameterBuilder".
            data_context: AbstractDataContext associated with this ParameterBuilder
        """
        super().__init__(
            name=name,
            metric_name="table.columns",
            metric_domain_kwargs=metric_domain_kwargs,
            metric_value_kwargs=metric_value_kwargs,
            enforce_numeric_metric=False,
            replace_nan_with_zero=False,
            reduce_scalar_metric=True,
            evaluation_parameter_builder_configs=evaluation_parameter_builder_configs,
            data_context=data_context,
        )

    def _build_parameters(
        self,
        domain: Domain,
        variables: Optional[ParameterContainer] = None,
        parameters: Optional[Dict[str, ParameterContainer]] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> Attributes:
        """
        Builds ParameterContainer object that holds ParameterNode objects with attribute name-value pairs and details.

        Returns:
            Attributes object, containing computed parameter values and parameter computation details metadata.
        """
        # Compute "table.columns" metric value for each Batch object.
        super().build_parameters(
            domain=domain,
            variables=variables,
            parameters=parameters,
            parameter_computation_impl=super()._build_parameters,
            runtime_configuration=runtime_configuration,
        )

        # Retrieve "table.columns" metric values for all Batch objects.
        parameter_node: ParameterNode = get_parameter_value_and_validate_return_type(
            domain=domain,
            parameter_reference=self.raw_fully_qualified_parameter_name,
            expected_return_type=None,
            variables=variables,
            parameters=parameters,
        )
        table_columns_names_multi_batch_value: MetricValues = parameter_node[
            FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY
        ]

        one_batch_table_columns_names_value: MetricValue
        multi_batch_table_columns_names_sets_as_list: List[Set[str]] = [
            set(one_batch_table_columns_names_value)  # type: ignore[arg-type] # could be dict
            for one_batch_table_columns_names_value in table_columns_names_multi_batch_value  # type: ignore[union-attr] # not all iterable
        ]

        multi_batch_table_columns_names_as_set: Set[str] = set().union(
            *multi_batch_table_columns_names_sets_as_list
        )

        one_batch_table_columns_names_set: Set[str]
        mean_table_columns_set_match: np.float64 = np.mean(
            np.asarray(
                [
                    1
                    if one_batch_table_columns_names_set
                    == multi_batch_table_columns_names_as_set
                    else 0
                    for one_batch_table_columns_names_set in multi_batch_table_columns_names_sets_as_list
                ]
            )
        )

        return Attributes(
            {
                FULLY_QUALIFIED_PARAMETER_NAME_VALUE_KEY: multi_batch_table_columns_names_as_set,
                FULLY_QUALIFIED_PARAMETER_NAME_METADATA_KEY: {
                    "success_ratio": mean_table_columns_set_match,
                },
            }
        )



import pandas as pd

from great_expectations.core.batch import RuntimeBatchRequest
from great_expectations.data_context import DataContext
from great_expectations.expectations.core.expect_column_mean_to_be_between import (
    ExpectColumnMeanToBeBetween,
)


# <snippet name="tests/expectations/core/test_expect_column_mean_to_be_positive.py ExpectColumnMeanToBePositive_class_def">
class ExpectColumnMeanToBePositive(ExpectColumnMeanToBeBetween):
    """Expects the mean of values in this column to be positive"""

    default_kwarg_values = {
        "min_value": 0,
        "strict_min": True,
    }
    # </snippet>
    # <snippet name="tests/expectations/core/test_expect_column_mean_to_be_positive.py validate_config">
    def validate_configuration(self, configuration):
        super().validate_configuration(configuration)
        assert "min_value" not in configuration.kwargs, "min_value cannot be altered"
        assert "max_value" not in configuration.kwargs, "max_value cannot be altered"
        assert "strict_min" not in configuration.kwargs, "strict_min cannot be altered"
        assert "strict_max" not in configuration.kwargs, "strict_max cannot be altered"

    # </snippet>
    # <snippet name="tests/expectations/core/test_expect_column_mean_to_be_positive.py library_metadata">
    library_metadata = {"tags": ["basic stats"], "contributors": ["@joegargery"]}


#     </snippet>


def test_expect_column_mean_to_be_positive(data_context_with_datasource_pandas_engine):
    context: DataContext = data_context_with_datasource_pandas_engine

    df = pd.DataFrame({"a": [0, 1, 3, 4, 5]})

    batch_request = RuntimeBatchRequest(
        datasource_name="my_datasource",
        data_connector_name="default_runtime_data_connector_name",
        data_asset_name="my_data_asset",
        runtime_parameters={"batch_data": df},
        batch_identifiers={"default_identifier_name": "my_identifier"},
    )
    validator = context.get_validator(
        batch_request=batch_request,
        create_expectation_suite_with_name="test",
    )

    result = validator.expect_column_mean_to_be_positive(column="a")

    assert result.success is True



import logging
import os
import random
from typing import Optional, Union

import pytest
from ruamel.yaml import YAML

from great_expectations.datasource import (
    BaseDatasource,
    LegacyDatasource,
    SimpleSqlalchemyDatasource,
)
from great_expectations.exceptions.exceptions import ExecutionEngineError

logger = logging.getLogger(__name__)


try:
    import pandas as pd
except ImportError:
    pd = None

    logger.debug(
        "Unable to load pandas; install optional pandas dependency for support."
    )

import great_expectations.exceptions as gx_exceptions
from great_expectations import DataContext
from great_expectations.core.batch import BatchRequest, RuntimeBatchRequest
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.data_context.util import (
    file_relative_path,
    instantiate_class_from_config,
)
from great_expectations.validator.validator import Validator

try:
    sqlalchemy = pytest.importorskip("sqlalchemy")
except ImportError:
    sqlalchemy = None

try:
    import sqlalchemy_bigquery as sqla_bigquery
except ImportError:
    try:
        import pybigquery.sqlalchemy_bigquery as sqla_bigquery
    except ImportError:
        sqla_bigquery = None

yaml = YAML()


@pytest.fixture
def data_context_with_sql_data_connectors_including_schema_for_testing_get_batch(
    sa,
    empty_data_context,
    test_db_connection_string,
):
    context: DataContext = empty_data_context

    sqlite_engine: sa.engine.base.Engine = sa.create_engine(test_db_connection_string)
    # noinspection PyUnusedLocal
    conn: sa.engine.base.Connection = sqlite_engine.connect()
    datasource_config: str = f"""
        class_name: Datasource

        execution_engine:
            class_name: SqlAlchemyExecutionEngine
            connection_string: {test_db_connection_string}

        data_connectors:
            my_runtime_data_connector:
                module_name: great_expectations.datasource.data_connector
                class_name: RuntimeDataConnector
                batch_identifiers:
                    - pipeline_stage_name
                    - airflow_run_id
            my_inferred_data_connector:
                module_name: great_expectations.datasource.data_connector
                class_name: InferredAssetSqlDataConnector
                include_schema_name: true
            my_configured_data_connector:
                module_name: great_expectations.datasource.data_connector
                class_name: ConfiguredAssetSqlDataConnector
                assets:
                    my_first_data_asset:
                        table_name: table_1
                    my_second_data_asset:
                        schema_name: main
                        table_name: table_2
                    table_1: {{}}
                    table_2:
                        include_schema_name: True
                        schema_name: main
    """

    try:
        # noinspection PyUnusedLocal
        my_sql_datasource: Optional[
            Union[SimpleSqlalchemyDatasource, LegacyDatasource]
        ] = context.add_datasource(
            "test_sqlite_db_datasource", **yaml.load(datasource_config)
        )
    except AttributeError:
        pytest.skip("SQL Database tests require sqlalchemy to be installed.")

    return context


@pytest.mark.integration
def test_basic_instantiation_with_ConfiguredAssetSqlDataConnector_splitting(sa):
    random.seed(0)

    db_file = file_relative_path(
        __file__,
        os.path.join("..", "test_sets", "test_cases_for_sql_data_connector.db"),
    )
    # This is a basic integration test demonstrating a Datasource containing a SQL data_connector
    # It also shows how to instantiate a SQLite SqlAlchemyExecutionEngine
    config = yaml.load(
        f"""
class_name: Datasource

execution_engine:
    class_name: SqlAlchemyExecutionEngine
    connection_string: sqlite:///{db_file}

data_connectors:
    my_sqlite_db:
        class_name: ConfiguredAssetSqlDataConnector

        assets:
            table_partitioned_by_date_column__A:
                splitter_method: _split_on_converted_datetime
                splitter_kwargs:
                    column_name: date
                    date_format_string: "%Y-%W"
    """,
    )

    my_data_source = instantiate_class_from_config(
        config,
        config_defaults={"module_name": "great_expectations.datasource"},
        runtime_environment={"name": "my_sql_datasource"},
    )

    report = my_data_source.self_check()
    report["execution_engine"].pop("connection_string")
    assert report == {
        "execution_engine": {
            "module_name": "great_expectations.execution_engine.sqlalchemy_execution_engine",
            "class_name": "SqlAlchemyExecutionEngine",
        },
        "data_connectors": {
            "count": 1,
            "my_sqlite_db": {
                "class_name": "ConfiguredAssetSqlDataConnector",
                "data_asset_count": 1,
                "example_data_asset_names": ["table_partitioned_by_date_column__A"],
                "data_assets": {
                    "table_partitioned_by_date_column__A": {
                        "batch_definition_count": 9,
                        "example_data_references": [
                            {"date": "2020-00"},
                            {"date": "2020-01"},
                            {"date": "2020-02"},
                        ],
                    }
                },
                "unmatched_data_reference_count": 0,
                "example_unmatched_data_references": [],
                # FIXME: (Sam) example_data_reference removed temporarily in PR #2590:
                # "example_data_reference": {
                #     "batch_spec": {
                #         "table_name": "table_partitioned_by_date_column__A",
                #         "data_asset_name": "table_partitioned_by_date_column__A",
                #         "batch_identifiers": {"date": "2020-01"},
                #         "splitter_method": "_split_on_converted_datetime",
                #         "splitter_kwargs": {
                #             "column_name": "date",
                #             "date_format_string": "%Y-%W",
                #         },
                #     },
                #     "n_rows": 24,
                # },
            },
        },
    }


@pytest.mark.integration
def test_instantiation_with_ConfiguredAssetSqlDataConnector_round_trip_to_config_splitting_and_sampling(
    sa, empty_data_context
):
    # This is a basic integration test demonstrating a Datasource containing a SQL data_connector.
    # It tests that splitter configurations can be saved and loaded to great_expectations.yml by performing a
    # round-trip to the configuration.
    context: DataContext = empty_data_context
    db_file: Union[bytes, str] = file_relative_path(
        __file__,
        os.path.join("..", "test_sets", "test_cases_for_sql_data_connector.db"),
    )
    config: str = f"""
    name: my_datasource
    class_name: Datasource

    execution_engine:
        class_name: SqlAlchemyExecutionEngine
        connection_string: sqlite:///{db_file}

    data_connectors:
        my_sqlite_db:
            class_name: ConfiguredAssetSqlDataConnector

            assets:
                table_partitioned_by_date_column__A:
                    splitter_method: _split_on_converted_datetime
                    splitter_kwargs:
                        column_name: date
                        date_format_string: "%Y-%W"
                    sampling_method: sample_using_limit
                    sampling_kwargs:
                        n: 10
    """
    context.add_datasource(**yaml.load(config))
    datasource: Union[LegacyDatasource, BaseDatasource, None] = context.get_datasource(
        datasource_name="my_datasource"
    )
    report: dict = datasource.self_check()
    report["execution_engine"].pop("connection_string")
    assert report == {
        "execution_engine": {
            "module_name": "great_expectations.execution_engine.sqlalchemy_execution_engine",
            "class_name": "SqlAlchemyExecutionEngine",
        },
        "data_connectors": {
            "count": 1,
            "my_sqlite_db": {
                "class_name": "ConfiguredAssetSqlDataConnector",
                "data_asset_count": 1,
                "example_data_asset_names": ["table_partitioned_by_date_column__A"],
                "data_assets": {
                    "table_partitioned_by_date_column__A": {
                        "batch_definition_count": 9,
                        "example_data_references": [
                            {"date": "2020-00"},
                            {"date": "2020-01"},
                            {"date": "2020-02"},
                        ],
                    }
                },
                "unmatched_data_reference_count": 0,
                "example_unmatched_data_references": [],
            },
        },
    }


@pytest.mark.integration
def test_basic_instantiation_with_InferredAssetSqlDataConnector_splitting(sa):
    # This is a basic integration test demonstrating a Datasource containing a SQL data_connector.
    # It tests that splitter configurations can be saved and loaded to great_expectations.yml by performing a
    # round-trip to the configuration.
    random.seed(0)

    db_file: Union[bytes, str] = file_relative_path(
        __file__,
        os.path.join("..", "test_sets", "test_cases_for_sql_data_connector.db"),
    )
    # This is a basic integration test demonstrating an Datasource containing a SQL data_connector
    # It also shows how to instantiate a SQLite SqlAlchemyExecutionEngine

    config = yaml.load(
        f"""
class_name: Datasource

execution_engine:
    class_name: SqlAlchemyExecutionEngine
    connection_string: sqlite:///{db_file}

data_connectors:
    my_sqlite_db:
        class_name: InferredAssetSqlDataConnector
        name: whole_table
        data_asset_name_prefix: prefix__
        data_asset_name_suffix: __xiffus
        splitter_method: _split_on_converted_datetime
        splitter_kwargs:
            column_name: date
            date_format_string: "%Y-%W"
    """,
    )

    my_data_source = instantiate_class_from_config(
        config,
        config_defaults={"module_name": "great_expectations.datasource"},
        runtime_environment={"name": "my_sql_datasource"},
    )
    report = my_data_source.self_check()
    report["execution_engine"].pop("connection_string")

    assert report == {
        "execution_engine": {
            "module_name": "great_expectations.execution_engine.sqlalchemy_execution_engine",
            "class_name": "SqlAlchemyExecutionEngine",
        },
        "data_connectors": {
            "count": 1,
            "my_sqlite_db": {
                "class_name": "InferredAssetSqlDataConnector",
                "data_asset_count": 6,
                "example_data_asset_names": [
                    "prefix__table_containing_id_spacers_for_D__xiffus",
                    "prefix__table_full__I__xiffus",
                    "prefix__table_partitioned_by_date_column__A__xiffus",
                ],
                "data_assets": {
                    "prefix__table_containing_id_spacers_for_D__xiffus": {
                        "batch_definition_count": 5,
                        "example_data_references": [
                            {"date": "2020-00"},
                            {"date": "2020-01"},
                            {"date": "2020-02"},
                        ],
                    },
                    "prefix__table_full__I__xiffus": {
                        "batch_definition_count": 5,
                        "example_data_references": [
                            {"date": "2020-00"},
                            {"date": "2020-01"},
                            {"date": "2020-02"},
                        ],
                    },
                    "prefix__table_partitioned_by_date_column__A__xiffus": {
                        "batch_definition_count": 9,
                        "example_data_references": [
                            {"date": "2020-00"},
                            {"date": "2020-01"},
                            {"date": "2020-02"},
                        ],
                    },
                },
                "unmatched_data_reference_count": 0,
                "example_unmatched_data_references": [],
            },
        },
    }


@pytest.mark.integration
def test_instantiation_with_InferredAssetSqlDataConnector_round_trip_to_config_splitting_and_sampling(
    sa, empty_data_context
):
    # This is a basic integration test demonstrating a Datasource containing a SQL data_connector.
    # It tests that splitter configurations can be saved and loaded to great_expectations.yml by performing a
    # round-trip to the configuration.
    context: DataContext = empty_data_context
    db_file: Union[bytes, str] = file_relative_path(
        __file__,
        os.path.join("..", "test_sets", "test_cases_for_sql_data_connector.db"),
    )
    config: str = f"""
    name: my_datasource
    class_name: Datasource

    execution_engine:
        class_name: SqlAlchemyExecutionEngine
        connection_string: sqlite:///{db_file}

    data_connectors:
        my_sqlite_db:
            class_name: InferredAssetSqlDataConnector
            name: whole_table
            data_asset_name_prefix: prefix__
            data_asset_name_suffix: __xiffus
            splitter_method: _split_on_converted_datetime
            splitter_kwargs:
                column_name: date
                date_format_string: "%Y-%W"
            sampling_method: sample_using_limit
            sampling_kwargs:
                n: 10
        """
    context.add_datasource(**yaml.load(config))
    datasource: Union[LegacyDatasource, BaseDatasource, None] = context.get_datasource(
        datasource_name="my_datasource"
    )
    report: dict = datasource.self_check()
    report["execution_engine"].pop("connection_string")
    assert report == {
        "execution_engine": {
            "module_name": "great_expectations.execution_engine.sqlalchemy_execution_engine",
            "class_name": "SqlAlchemyExecutionEngine",
        },
        "data_connectors": {
            "count": 1,
            "my_sqlite_db": {
                "class_name": "InferredAssetSqlDataConnector",
                "data_asset_count": 6,
                "example_data_asset_names": [
                    "prefix__table_containing_id_spacers_for_D__xiffus",
                    "prefix__table_full__I__xiffus",
                    "prefix__table_partitioned_by_date_column__A__xiffus",
                ],
                "data_assets": {
                    "prefix__table_containing_id_spacers_for_D__xiffus": {
                        "batch_definition_count": 5,
                        "example_data_references": [
                            {"date": "2020-00"},
                            {"date": "2020-01"},
                            {"date": "2020-02"},
                        ],
                    },
                    "prefix__table_full__I__xiffus": {
                        "batch_definition_count": 5,
                        "example_data_references": [
                            {"date": "2020-00"},
                            {"date": "2020-01"},
                            {"date": "2020-02"},
                        ],
                    },
                    "prefix__table_partitioned_by_date_column__A__xiffus": {
                        "batch_definition_count": 9,
                        "example_data_references": [
                            {"date": "2020-00"},
                            {"date": "2020-01"},
                            {"date": "2020-02"},
                        ],
                    },
                },
                "unmatched_data_reference_count": 0,
                "example_unmatched_data_references": [],
            },
        },
    }


@pytest.mark.integration
def test_SimpleSqlalchemyDatasource(empty_data_context):
    context = empty_data_context
    # This test mirrors the likely path to configure a SimpleSqlalchemyDatasource

    db_file = file_relative_path(
        __file__,
        os.path.join("..", "test_sets", "test_cases_for_sql_data_connector.db"),
    )

    # Absolutely minimal starting config
    datasource_with_minimum_config = context.test_yaml_config(
        f"""
class_name: SimpleSqlalchemyDatasource
connection_string: sqlite:///{db_file}
"""
        + """
introspection:
    whole_table: {}
"""
    )

    assert datasource_with_minimum_config.get_available_data_asset_names() == {
        "whole_table": [
            "table_containing_id_spacers_for_D",
            "table_full__I",
            "table_partitioned_by_date_column__A",
            "table_partitioned_by_foreign_key__F",
            "table_partitioned_by_incrementing_batch_id__E",
            "table_partitioned_by_irregularly_spaced_incrementing_id_with_spacing_in_a_second_table__D",
            "table_partitioned_by_multiple_columns__G",
            "table_partitioned_by_regularly_spaced_incrementing_id_column__C",
            "table_partitioned_by_timestamp_column__B",
            "table_that_should_be_partitioned_by_random_hash__H",
            "table_with_fk_reference_from_F",
            "view_by_date_column__A",
            "view_by_incrementing_batch_id__E",
            "view_by_irregularly_spaced_incrementing_id_with_spacing_in_a_second_table__D",
            "view_by_multiple_columns__G",
            "view_by_regularly_spaced_incrementing_id_column__C",
            "view_by_timestamp_column__B",
            "view_containing_id_spacers_for_D",
            "view_partitioned_by_foreign_key__F",
            "view_that_should_be_partitioned_by_random_hash__H",
            "view_with_fk_reference_from_F",
        ]
    }

    assert datasource_with_minimum_config.get_available_data_asset_names_and_types() == {
        "whole_table": [
            ("table_containing_id_spacers_for_D", "table"),
            ("table_full__I", "table"),
            ("table_partitioned_by_date_column__A", "table"),
            ("table_partitioned_by_foreign_key__F", "table"),
            ("table_partitioned_by_incrementing_batch_id__E", "table"),
            (
                "table_partitioned_by_irregularly_spaced_incrementing_id_with_spacing_in_a_second_table__D",
                "table",
            ),
            ("table_partitioned_by_multiple_columns__G", "table"),
            (
                "table_partitioned_by_regularly_spaced_incrementing_id_column__C",
                "table",
            ),
            ("table_partitioned_by_timestamp_column__B", "table"),
            ("table_that_should_be_partitioned_by_random_hash__H", "table"),
            ("table_with_fk_reference_from_F", "table"),
            ("view_by_date_column__A", "view"),
            ("view_by_incrementing_batch_id__E", "view"),
            (
                "view_by_irregularly_spaced_incrementing_id_with_spacing_in_a_second_table__D",
                "view",
            ),
            ("view_by_multiple_columns__G", "view"),
            ("view_by_regularly_spaced_incrementing_id_column__C", "view"),
            ("view_by_timestamp_column__B", "view"),
            ("view_containing_id_spacers_for_D", "view"),
            ("view_partitioned_by_foreign_key__F", "view"),
            ("view_that_should_be_partitioned_by_random_hash__H", "view"),
            ("view_with_fk_reference_from_F", "view"),
        ]
    }

    # Here we should test getting a batch

    # Very thin starting config
    datasource_with_name_suffix = context.test_yaml_config(
        f"""
class_name: SimpleSqlalchemyDatasource
connection_string: sqlite:///{db_file}
"""
        + """
introspection:
    whole_table:
        data_asset_name_suffix: __whole_table
        introspection_directives: {}
"""
    )

    assert datasource_with_name_suffix.get_available_data_asset_names() == {
        "whole_table": [
            "table_containing_id_spacers_for_D__whole_table",
            "table_full__I__whole_table",
            "table_partitioned_by_date_column__A__whole_table",
            "table_partitioned_by_foreign_key__F__whole_table",
            "table_partitioned_by_incrementing_batch_id__E__whole_table",
            "table_partitioned_by_irregularly_spaced_incrementing_id_with_spacing_in_a_second_table__D__whole_table",
            "table_partitioned_by_multiple_columns__G__whole_table",
            "table_partitioned_by_regularly_spaced_incrementing_id_column__C__whole_table",
            "table_partitioned_by_timestamp_column__B__whole_table",
            "table_that_should_be_partitioned_by_random_hash__H__whole_table",
            "table_with_fk_reference_from_F__whole_table",
            "view_by_date_column__A__whole_table",
            "view_by_incrementing_batch_id__E__whole_table",
            "view_by_irregularly_spaced_incrementing_id_with_spacing_in_a_second_table__D__whole_table",
            "view_by_multiple_columns__G__whole_table",
            "view_by_regularly_spaced_incrementing_id_column__C__whole_table",
            "view_by_timestamp_column__B__whole_table",
            "view_containing_id_spacers_for_D__whole_table",
            "view_partitioned_by_foreign_key__F__whole_table",
            "view_that_should_be_partitioned_by_random_hash__H__whole_table",
            "view_with_fk_reference_from_F__whole_table",
        ]
    }

    assert datasource_with_name_suffix.get_available_data_asset_names_and_types() == {
        "whole_table": [
            ("table_containing_id_spacers_for_D", "table"),
            ("table_full__I", "table"),
            ("table_partitioned_by_date_column__A", "table"),
            ("table_partitioned_by_foreign_key__F", "table"),
            ("table_partitioned_by_incrementing_batch_id__E", "table"),
            (
                "table_partitioned_by_irregularly_spaced_incrementing_id_with_spacing_in_a_second_table__D",
                "table",
            ),
            ("table_partitioned_by_multiple_columns__G", "table"),
            (
                "table_partitioned_by_regularly_spaced_incrementing_id_column__C",
                "table",
            ),
            ("table_partitioned_by_timestamp_column__B", "table"),
            ("table_that_should_be_partitioned_by_random_hash__H", "table"),
            ("table_with_fk_reference_from_F", "table"),
            ("view_by_date_column__A", "view"),
            ("view_by_incrementing_batch_id__E", "view"),
            (
                "view_by_irregularly_spaced_incrementing_id_with_spacing_in_a_second_table__D",
                "view",
            ),
            ("view_by_multiple_columns__G", "view"),
            ("view_by_regularly_spaced_incrementing_id_column__C", "view"),
            ("view_by_timestamp_column__B", "view"),
            ("view_containing_id_spacers_for_D", "view"),
            ("view_partitioned_by_foreign_key__F", "view"),
            ("view_that_should_be_partitioned_by_random_hash__H", "view"),
            ("view_with_fk_reference_from_F", "view"),
        ]
    }

    # Here we should test getting a batch

    # Add some manually configured tables...
    datasource_manually_configured = context.test_yaml_config(
        f"""
class_name: SimpleSqlalchemyDatasource
connection_string: sqlite:///{db_file}

introspection:
    whole_table:
        excluded_tables:
            - main.table_partitioned_by_irregularly_spaced_incrementing_id_with_spacing_in_a_second_table__D
            - main.table_partitioned_by_multiple_columns__G
            - main.table_partitioned_by_regularly_spaced_incrementing_id_column__C
            - main.table_partitioned_by_timestamp_column__B
            - main.table_that_should_be_partitioned_by_random_hash__H
            - main.table_with_fk_reference_from_F

    hourly:
        splitter_method: _split_on_converted_datetime
        splitter_kwargs:
            column_name: timestamp
            date_format_string: "%Y-%m-%d:%H"
        included_tables:
            - main.table_partitioned_by_timestamp_column__B
        introspection_directives:
            include_views: true


tables:
    table_partitioned_by_date_column__A:
        partitioners:
            daily:
                data_asset_name_suffix: __daily
                splitter_method: _split_on_converted_datetime
                splitter_kwargs:
                    column_name: date
                    date_format_string: "%Y-%m-%d"
            weekly:
                include_schema_name: False
                data_asset_name_prefix: some_string__
                data_asset_name_suffix: __some_other_string
                splitter_method: _split_on_converted_datetime
                splitter_kwargs:
                    column_name: date
                    date_format_string: "%Y-%W"
            by_id_dozens:
                include_schema_name: True
                # Note: no data_asset_name_suffix
                splitter_method: _split_on_divided_integer
                splitter_kwargs:
                    column_name: id
                    divisor: 12
"""
    )

    assert datasource_manually_configured.get_available_data_asset_names() == {
        "whole_table": [
            "table_containing_id_spacers_for_D",
            "table_full__I",
            "table_partitioned_by_date_column__A",
            "table_partitioned_by_foreign_key__F",
            "table_partitioned_by_incrementing_batch_id__E",
            "view_by_date_column__A",
            "view_by_incrementing_batch_id__E",
            "view_by_irregularly_spaced_incrementing_id_with_spacing_in_a_second_table__D",
            "view_by_multiple_columns__G",
            "view_by_regularly_spaced_incrementing_id_column__C",
            "view_by_timestamp_column__B",
            "view_containing_id_spacers_for_D",
            "view_partitioned_by_foreign_key__F",
            "view_that_should_be_partitioned_by_random_hash__H",
            "view_with_fk_reference_from_F",
        ],
        "hourly": [
            "table_partitioned_by_timestamp_column__B",
        ],
        "daily": [
            "table_partitioned_by_date_column__A__daily",
        ],
        "weekly": [
            "some_string__table_partitioned_by_date_column__A__some_other_string",
        ],
        "by_id_dozens": [
            "table_partitioned_by_date_column__A",
        ],
    }

    # Here we should test getting another batch

    # Drop the introspection...
    datasource_without_introspection = context.test_yaml_config(
        f"""
class_name: SimpleSqlalchemyDatasource
connection_string: sqlite:///{db_file}
"""
        + """
tables:
    table_partitioned_by_date_column__A:
        partitioners:
            whole_table: {}
            daily:
                splitter_method: _split_on_converted_datetime
                splitter_kwargs:
                    column_name: date
                    date_format_string: "%Y-%m-%d"
            weekly:
                splitter_method: _split_on_converted_datetime
                splitter_kwargs:
                    column_name: date
                    date_format_string: "%Y-%W"
            by_id_dozens:
                splitter_method: _split_on_divided_integer
                splitter_kwargs:
                    column_name: id
                    divisor: 12
"""
    )

    assert datasource_without_introspection.get_available_data_asset_names() == {
        "whole_table": [
            "table_partitioned_by_date_column__A",
        ],
        "daily": [
            "table_partitioned_by_date_column__A",
        ],
        "weekly": [
            "table_partitioned_by_date_column__A",
        ],
        "by_id_dozens": [
            "table_partitioned_by_date_column__A",
        ],
    }

    # Here we should test getting another batch


@pytest.mark.skipif(
    sqla_bigquery is None,
    reason="sqlalchemy_bigquery/pybigquery is not installed",
)
@pytest.mark.integration
def test_basic_instantiation_with_bigquery_creds(sa):
    # bigquery driver is invoked upon datasource instantiation, and validates credentials_info
    instantiate_class_from_config(
        # private key is valid but useless
        config={
            "connection_string": "bigquery://project-1353/dataset",
            "credentials_info": {
                "type": "service_account",
                "project_id": "project-1353",
                "private_key_id": "df87033061fd7c27dcc953e235fe099a7017f9c4",
                "private_key": "-----BEGIN PRIVATE KEY-----\nMIIEvAIBADANBgkqhkiG9w0BAQEFAASCBKYwggSiAgEAAoIBAQDJGtBAt//4Mro3\n58HfA8JfoX1KtTtCCsvGqVhXY5Z74q7knLJUm3cy6pheIUrJ7qvTSuzQZrguI+x5\n/J3+iosjgEhtkHfkDwPreW/K5TL/JXh1xjr/qMro4P10csVmmiAtVDypFIj2Tl4f\n6oJ5qGJJ6WBtFmswgHZeSZCis3FYIpho9Dqzj/Uahji4/ApaY4s/O463ymr4VhFW\n3ieIiepUB/XBBYzJOa5xAmo5HA6aSw7SJX+HUdxqHyhdsIz6kNV+3ej/YF4hgzVF\n2vn2+/+71GqSaZKK12frdfFU4VkWMzgvr2tO9XYX49tPOHTlZdEmi1UV2VuS2M39\nsuHHpaOTAgMBAAECggEARRGg+cFYN/nQLDg8RSiI3whbPEffSNDlaN8rmKP7AKR7\ntce9lcJpX4Lj/txHT/BZcjG3AOJummY7JzBkYRJbND+wYHTwQFMJ4Rttkk1CxQ+s\n/iItjDYALphrZE2wz4rax0a5qMaFPbbvq92Cn17+Fu2A8SZ0fQ152etBMigYIxDu\nzvmT4Fb67zJF5BOt8ay75a50H2sxtJcOWiaFX4Esil7+9gJft04MHSsZXBh8GgQK\nKo8xFlBe37tT1vE6Np0igxJIm+HjqObOpQ0pkaE1H2/rWt7k+HfuEWxKYva2t8C2\neexSupeTj1AkwgzKhk7sbyKMieTWg8+Tc3UODxZEMQKBgQDm4Slc8g5e59xw/Wg2\neyM/Gv2Syh1NLDChuFph4A0SBZhNYG7v5BZE4hVIHfZrrLzqSq7KQ2RojDIQzB7u\nUnrtdblvjZcwc1eepgbB9yjcUifaANZg8k9ukN2V0glPvYtIE1yHX6p8kvrIOlEb\n+WxTWXecmq8FL5QnUvboByFogwKBgQDe/FCneKBmaQPEJWa1cx2izA3JFyViRKx5\nWMv3CqEAL96YuL2JOnEvZFGeG+Bup8nohA8YD58TaED58vKjs9j0WFxw9T2V5d9M\n24fT0AwcZMnnrMKslC6ShyWMzntqHL3FvLh4jBthcJ+fsJbLN+vs/7qhU0qWht26\nvCzqAPpLsQKBgAmmMWdcCnO29wSV4qwcO13gz+Y5oj3ece3gWY6roYA2UaYyOJC4\nFKIuXLtV3T2ky4RzOJjldiXUbic7kLNfKRoRiH18CmyQ9YGA6NlkbgW/PUEkNdF5\nbw5s6YXgcFkvz8lkXcKeoe5w6iBCJ6+mnfthytj1sgjicyutkPojiibnAoGAKBEu\nbOk/6Sb1hkkyK2iD6ry/vWJrVT0BwMwz2jAOvfncBZUseXwG2n0sLTzVFw0POrh/\n/dLQwqv5APCmqMOoOD+oXKO0bTrg5O6NeYHoqzFxFi/0yw3VUH74QFTZ2DdR4jYG\n76I9SUTzab5RWjKyMePBpTtSK7oQHX2ylFmYoAECgYBmkwKQkj1b0WZ3+lbb79Jc\nZayAGUgg/W1Fh0V1m8on8wzGOoBoYSbmriyUlUycEivnJaskxCU1Yac2hNQS8KaU\nnm/gd3D0/8ghNW9szvvvRKc99dpgU6OYHvESq4+vG5gdIDHm2C5jMQGQgf1l2VOV\n0z7hg6e3jgecJweN7Yzfnw==\n-----END PRIVATE KEY-----\n",
                "client_email": "testme@project-1353.iam.gserviceaccount.com",
                "client_id": "100945395817716260007",
                "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                "token_uri": "https://oauth2.googleapis.com/token",
                "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/testme%40project-1353.iam.gserviceaccount.com",
            },
        },
        config_defaults={
            "module_name": "great_expectations.datasource",
            "class_name": "SimpleSqlalchemyDatasource",
        },
        runtime_environment={"name": "my_sql_datasource"},
    )


@pytest.mark.integration
def test_basic_instantiation_with_bigquery_creds_failure_pkey(sa):
    with pytest.raises(ExecutionEngineError):
        # private key is valid but useless so an exception should be raised.
        instantiate_class_from_config(
            config={
                "connection_string": "bigquery://project-1353/dataset",
                "credentials_info": {
                    "type": "service_account",
                    "project_id": "project-1353",
                    "private_key_id": "df87033061fd7c27dcc953e235fe099a7017f9c4",
                    "private_key": "bad_pkey",
                    "client_email": "testme@project-1353.iam.gserviceaccount.com",
                    "client_id": "100945395817716260007",
                    "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                    "token_uri": "https://oauth2.googleapis.com/token",
                    "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                    "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/testme%40project-1353.iam.gserviceaccount.com",
                },
            },
            config_defaults={
                "module_name": "great_expectations.datasource",
                "class_name": "SimpleSqlalchemyDatasource",
            },
            runtime_environment={"name": "my_sql_datasource"},
        )


@pytest.mark.integration
def test_skip_inapplicable_tables(empty_data_context):
    context = empty_data_context
    # This test mirrors the likely path to configure a SimpleSqlalchemyDatasource

    db_file = file_relative_path(
        __file__,
        os.path.join("..", "test_sets", "test_cases_for_sql_data_connector.db"),
    )

    my_sql_datasource = context.test_yaml_config(
        f"""
class_name: SimpleSqlalchemyDatasource
connection_string: sqlite:///{db_file}
introspection:
    daily:
        skip_inapplicable_tables: true
        splitter_method: _split_on_converted_datetime
        splitter_kwargs:
            column_name: date
            date_format_string: "%Y-%m-%d"
"""
    )
    assert my_sql_datasource.get_available_data_asset_names() == {
        "daily": [
            "table_containing_id_spacers_for_D",
            "table_full__I",
            "table_partitioned_by_date_column__A",
            "table_with_fk_reference_from_F",
            "view_by_date_column__A",
            "view_with_fk_reference_from_F",
        ]
    }

    with pytest.raises(gx_exceptions.DatasourceInitializationError):
        # noinspection PyUnusedLocal
        my_sql_datasource = context.test_yaml_config(
            f"""
class_name: SimpleSqlalchemyDatasource
connection_string: sqlite:///{db_file}
introspection:
    daily:
        skip_inapplicable_tables: false
        splitter_method: _split_on_converted_datetime
        splitter_kwargs:
            column_name: date
            date_format_string: "%Y-%m-%d"
    """
        )


@pytest.mark.integration
def test_batch_request_sql_with_schema(
    data_context_with_sql_data_connectors_including_schema_for_testing_get_batch,
):
    context: DataContext = (
        data_context_with_sql_data_connectors_including_schema_for_testing_get_batch
    )

    df_table_expected_my_first_data_asset: pd.DataFrame = pd.DataFrame(
        {"col_1": [1, 2, 3, 4, 5], "col_2": ["a", "b", "c", "d", "e"]}
    )
    df_table_expected_my_second_data_asset: pd.DataFrame = pd.DataFrame(
        {"col_1": [0, 1, 2, 3, 4], "col_2": ["b", "c", "d", "e", "f"]}
    )

    batch_request: dict
    validator: Validator
    df_table_actual: pd.DataFrame

    # Exercise RuntimeDataConnector using SQL query against database table with empty schema name.
    batch_request = {
        "datasource_name": "test_sqlite_db_datasource",
        "data_connector_name": "my_runtime_data_connector",
        "data_asset_name": "test_asset",
        "runtime_parameters": {"query": "SELECT * FROM table_1"},
        "batch_identifiers": {
            "pipeline_stage_name": "core_processing",
            "airflow_run_id": 1234567890,
        },
    }
    validator = context.get_validator(
        batch_request=RuntimeBatchRequest(**batch_request),
        expectation_suite=ExpectationSuite(
            "my_expectation_suite", data_context=context
        ),
    )
    df_table_actual = validator.head(n_rows=0, fetch_all=True).drop(columns=["index"])
    assert df_table_actual.equals(df_table_expected_my_first_data_asset)

    # Exercise RuntimeDataConnector using SQL query against database table with non-empty ("main") schema name.
    batch_request = {
        "datasource_name": "test_sqlite_db_datasource",
        "data_connector_name": "my_runtime_data_connector",
        "data_asset_name": "test_asset",
        "runtime_parameters": {"query": "SELECT * FROM main.table_2"},
        "batch_identifiers": {
            "pipeline_stage_name": "core_processing",
            "airflow_run_id": 1234567891,
        },
    }
    validator = context.get_validator(
        batch_request=RuntimeBatchRequest(**batch_request),
        expectation_suite=ExpectationSuite(
            "my_expectation_suite", data_context=context
        ),
    )
    df_table_actual = validator.head(n_rows=0, fetch_all=True).drop(columns=["index"])
    assert df_table_actual.equals(df_table_expected_my_second_data_asset)

    # Exercise InferredAssetSqlDataConnector using data_asset_name introspected with schema from table, named "table_1".
    batch_request = {
        "datasource_name": "test_sqlite_db_datasource",
        "data_connector_name": "my_inferred_data_connector",
        "data_asset_name": "main.table_1",
    }
    validator = context.get_validator(
        batch_request=BatchRequest(**batch_request),
        expectation_suite=ExpectationSuite(
            "my_expectation_suite", data_context=context
        ),
    )
    df_table_actual = validator.head(n_rows=0, fetch_all=True).drop(columns=["index"])
    assert df_table_actual.equals(df_table_expected_my_first_data_asset)

    # Exercise InferredAssetSqlDataConnector using data_asset_name introspected with schema from table, named "table_2".
    batch_request = {
        "datasource_name": "test_sqlite_db_datasource",
        "data_connector_name": "my_inferred_data_connector",
        "data_asset_name": "main.table_2",
    }
    validator = context.get_validator(
        batch_request=BatchRequest(**batch_request),
        expectation_suite=ExpectationSuite(
            "my_expectation_suite", data_context=context
        ),
    )
    df_table_actual = validator.head(n_rows=0, fetch_all=True).drop(columns=["index"])
    assert df_table_actual.equals(df_table_expected_my_second_data_asset)

    # Exercise ConfiguredAssetSqlDataConnector using data_asset_name corresponding to "my_first_data_asset" (implicitly).
    batch_request = {
        "datasource_name": "test_sqlite_db_datasource",
        "data_connector_name": "my_configured_data_connector",
        "data_asset_name": "my_first_data_asset",
    }
    validator = context.get_validator(
        batch_request=BatchRequest(**batch_request),
        expectation_suite=ExpectationSuite(
            "my_expectation_suite", data_context=context
        ),
    )
    df_table_actual = validator.head(n_rows=0, fetch_all=True).drop(columns=["index"])
    assert df_table_actual.equals(df_table_expected_my_first_data_asset)

    # Exercise ConfiguredAssetSqlDataConnector using data_asset_name corresponding to "my_second_data_asset" (implicitly).
    batch_request = {
        "datasource_name": "test_sqlite_db_datasource",
        "data_connector_name": "my_configured_data_connector",
        "data_asset_name": "my_second_data_asset",
    }
    validator = context.get_validator(
        batch_request=BatchRequest(**batch_request),
        expectation_suite=ExpectationSuite(
            "my_expectation_suite", data_context=context
        ),
    )
    df_table_actual = validator.head(n_rows=0, fetch_all=True).drop(columns=["index"])
    assert df_table_actual.equals(df_table_expected_my_second_data_asset)

    # Exercise ConfiguredAssetSqlDataConnector using data_asset_name corresponding to "table_1" (explicitly).
    batch_request = {
        "datasource_name": "test_sqlite_db_datasource",
        "data_connector_name": "my_configured_data_connector",
        "data_asset_name": "table_1",
    }
    validator = context.get_validator(
        batch_request=BatchRequest(**batch_request),
        expectation_suite=ExpectationSuite(
            "my_expectation_suite", data_context=context
        ),
    )
    df_table_actual = validator.head(n_rows=0, fetch_all=True).drop(columns=["index"])
    assert df_table_actual.equals(df_table_expected_my_first_data_asset)

    # Exercise ConfiguredAssetSqlDataConnector using data_asset_name corresponding to "main.table_2" (explicitly).
    batch_request = {
        "datasource_name": "test_sqlite_db_datasource",
        "data_connector_name": "my_configured_data_connector",
        "data_asset_name": "main.table_2",
    }
    validator = context.get_validator(
        batch_request=BatchRequest(**batch_request),
        expectation_suite=ExpectationSuite(
            "my_expectation_suite", data_context=context
        ),
    )
    df_table_actual = validator.head(n_rows=0, fetch_all=True).drop(columns=["index"])
    assert df_table_actual.equals(df_table_expected_my_second_data_asset)



import logging
from abc import ABC, abstractmethod
from hashlib import md5
from typing import Any, List, Optional, Tuple

from great_expectations.core.usage_statistics.util import (
    aggregate_all_core_expectation_types,
)
from great_expectations.util import load_class

logger = logging.getLogger(__name__)


class BaseAnonymizer(ABC):

    # Any class that starts with this __module__ is considered a "core" object
    CORE_GX_OBJECT_MODULE_PREFIX = "great_expectations"

    CORE_GX_EXPECTATION_TYPES = aggregate_all_core_expectation_types()

    def __init__(self, salt: Optional[str] = None) -> None:
        if salt is not None and not isinstance(salt, str):
            logger.error("invalid salt: must provide a string. Setting a random salt.")
            salt = None
        if salt is None:
            import secrets

            self._salt = secrets.token_hex(8)
        else:
            self._salt = salt

    @abstractmethod
    def anonymize(self, obj: Optional[object], **kwargs) -> Any:
        raise NotImplementedError

    @abstractmethod
    def can_handle(self, obj: Optional[object], **kwargs) -> bool:
        raise NotImplementedError

    @staticmethod
    def get_parent_class(
        classes_to_check: Optional[List[type]] = None,
        object_: Optional[object] = None,
        object_class: Optional[type] = None,
        object_config: Optional[dict] = None,
    ) -> Optional[str]:
        """Check if the parent class is a subclass of any core GX class.

        These anonymizers define and provide an optional list of core GX classes_to_check.
        If not provided, the object's inheritance hierarchy is traversed.

        Args:
            classes_to_check: An optinal list of candidate parent classes to iterate through.
            object_: The specific object to analyze.
            object_class: The class of the specific object to analyze.
            object_config: The dictionary configuration of the specific object to analyze.

        Returns:
            The name of the parent class found, or None if no parent class was found.

        Raises:
            AssertionError if no object_, object_class, or object_config is provided.
        """
        assert (
            object_ or object_class or object_config
        ), "Must pass either object_ or object_class or object_config."
        try:
            if object_class is None and object_ is not None:
                object_class = object_.__class__
            elif object_class is None and object_config is not None:
                object_class_name = object_config.get("class_name")
                object_module_name = object_config.get("module_name")
                object_class = load_class(object_class_name, object_module_name)

            # Utilize candidate list if provided.
            if classes_to_check:
                for class_to_check in classes_to_check:
                    if issubclass(object_class, class_to_check):
                        return class_to_check.__name__
                return None

            # Otherwise, iterate through parents in inheritance hierarchy.
            parents: Tuple[type, ...] = object_class.__bases__
            parent_class: type
            for parent_class in parents:
                parent_module_name: str = parent_class.__module__
                if BaseAnonymizer._is_core_great_expectations_class(parent_module_name):
                    return parent_class.__name__

        except AttributeError:
            pass

        return None

    def _anonymize_string(self, string_: Optional[str]) -> Optional[str]:
        """Obsfuscates a given string using an MD5 hash.

        Utilized to anonymize user-specific/sensitive strings in usage statistics payloads.

        Args:
            string_ (Optional[str]): The input string to anonymize.

        Returns:
            The MD5 hash of the input string. If an input of None is provided, the string is untouched.

        Raises:
            TypeError if input string does not adhere to type signature Optional[str].
        """
        if string_ is None:
            return None

        if not isinstance(string_, str):
            raise TypeError(
                f"""The type of the "string_" argument must be a string (Python "str").  The type given is
"{str(type(string_))}", which is illegal.
            """
            )

        salted = self._salt + string_
        return md5(salted.encode("utf-8")).hexdigest()

    def _anonymize_object_info(
        self,
        anonymized_info_dict: dict,
        object_: Optional[object] = None,
        object_class: Optional[type] = None,
        object_config: Optional[dict] = None,
        runtime_environment: Optional[dict] = None,
    ) -> dict:
        """Given an object, anonymize relevant fields and return result as a dictionary.

        Args:
            anonymized_info_dict: The payload object to hydrate with anonymized values.
            object_: The specific object to anonymize.
            object_class: The class of the specific object to anonymize.
            object_config: The dictionary configuration of the specific object to anonymize.
            runtime_environment: A dictionary containing relevant runtime information (like class_name and module_name)

        Returns:
            The anonymized_info_dict that's been populated with anonymized values.

        Raises:
            AssertionError if no object_, object_class, or object_config is provided.
        """
        assert (
            object_ or object_class or object_config
        ), "Must pass either object_ or object_class or object_config."

        if runtime_environment is None:
            runtime_environment = {}

        object_class_name: Optional[str] = None
        object_module_name: Optional[str] = None
        try:
            if object_class is None and object_ is not None:
                object_class = object_.__class__
            elif object_class is None and object_config is not None:
                object_class_name = object_config.get("class_name")
                object_module_name = object_config.get(
                    "module_name"
                ) or runtime_environment.get("module_name")
                object_class = load_class(object_class_name, object_module_name)

            object_class_name = object_class.__name__
            object_module_name = object_class.__module__
            parents: Tuple[type, ...] = object_class.__bases__

            if self._is_core_great_expectations_class(object_module_name):
                anonymized_info_dict["parent_class"] = object_class_name
            else:

                # Chetan - 20220311 - If we can't identify the class in question, we iterate through the parents.
                # While GX rarely utilizes multiple inheritance when defining core objects (as of v0.14.10),
                # it is important to recognize that this is possibility.
                #
                # In the presence of multiple valid parents, we generate a comma-delimited list.

                parent_class_list: List[str] = []

                parent_class: type
                for parent_class in parents:
                    parent_module_name: str = parent_class.__module__
                    if BaseAnonymizer._is_core_great_expectations_class(
                        parent_module_name
                    ):
                        parent_class_list.append(parent_class.__name__)

                if parent_class_list:
                    concatenated_parent_classes: str = ",".join(
                        cls for cls in parent_class_list
                    )
                    anonymized_info_dict["parent_class"] = concatenated_parent_classes
                    anonymized_info_dict["anonymized_class"] = self._anonymize_string(
                        object_class_name
                    )

            # Catch-all to prevent edge cases from slipping past
            if not anonymized_info_dict.get("parent_class"):
                anonymized_info_dict["parent_class"] = "__not_recognized__"
                anonymized_info_dict["anonymized_class"] = self._anonymize_string(
                    object_class_name
                )

        except AttributeError:
            anonymized_info_dict["parent_class"] = "__not_recognized__"
            anonymized_info_dict["anonymized_class"] = self._anonymize_string(
                object_class_name
            )

        return anonymized_info_dict

    @staticmethod
    def _is_core_great_expectations_class(class_name: str) -> bool:
        return class_name.startswith(BaseAnonymizer.CORE_GX_OBJECT_MODULE_PREFIX)
